from unittest.mock import AsyncMock, Mock, patch

import pytest

from hypertic.vectordb.base import VectorDocument, VectorSearchResult
from hypertic.vectordb.chroma.chroma import ChromaDB
from hypertic.vectordb.mongovector.mongovector import MongoDBAtlas
from hypertic.vectordb.pgvector.pgvector import PgVectorDB
from hypertic.vectordb.pinecone.pinecone import PineconeDB
from hypertic.vectordb.qdrant.qdrant import QdrantDB


class TestVectorDBIntegration:
    @pytest.mark.parametrize(
        "db_class,db_params",
        [
            (ChromaDB, {"embedder": Mock(), "collection": "test"}),
            (PineconeDB, {"api_key": "test", "embedder": Mock(), "collection": "test"}),
            (QdrantDB, {"embedder": Mock(), "collection": "test"}),
            (PgVectorDB, {"embedder": Mock(), "db_url": "postgresql://localhost/test", "collection": "test"}),
            (MongoDBAtlas, {"embedder": Mock(), "connection_string": "mongodb://localhost:27017", "database_name": "test", "collection": "test"}),
        ],
    )
    def test_vectordb_initialization(self, db_class, db_params):
        db = db_class(**db_params)
        assert db is not None

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "db_class,db_params",
        [
            (ChromaDB, {"embedder": Mock(), "collection": "test"}),
            (PineconeDB, {"api_key": "test", "embedder": Mock(), "collection": "test"}),
            (QdrantDB, {"embedder": Mock(), "collection": "test"}),
            (PgVectorDB, {"embedder": Mock(), "db_url": "postgresql://localhost/test", "collection": "test"}),
            (MongoDBAtlas, {"embedder": Mock(), "connection_string": "mongodb://localhost:27017", "database_name": "test", "collection": "test"}),
        ],
    )
    async def test_vectordb_async_initialize(self, db_class, db_params):
        db = db_class(**db_params)

        with patch.object(db, "initialize", return_value=True):
            result = await db.initialize()
            assert result is True

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "db_class,db_params",
        [
            (ChromaDB, {"embedder": Mock(), "collection": "test"}),
            (PineconeDB, {"api_key": "test", "embedder": Mock(), "collection": "test"}),
            (QdrantDB, {"embedder": Mock(), "collection": "test"}),
            (PgVectorDB, {"embedder": Mock(), "db_url": "postgresql://localhost/test", "collection": "test"}),
            (MongoDBAtlas, {"embedder": Mock(), "connection_string": "mongodb://localhost:27017", "database_name": "test", "collection": "test"}),
        ],
    )
    async def test_vectordb_async_add_documents(self, db_class, db_params, mock_embedding):
        db = db_class(**db_params)

        documents = [
            VectorDocument(
                id="doc_1",
                content="Test content",
                vector=mock_embedding,
                metadata={"source": "test"},
            )
        ]

        if hasattr(db, "add_documents_async"):
            with patch.object(db, "add_documents_async", return_value=None):
                await db.add_documents_async(documents)
        else:
            with patch.object(db, "add_documents", return_value=None):
                await db.add_documents(documents)
            assert True

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "db_class,db_params",
        [
            (ChromaDB, {"embedder": Mock(), "collection": "test"}),
            (PineconeDB, {"api_key": "test", "embedder": Mock(), "collection": "test"}),
            (QdrantDB, {"embedder": Mock(), "collection": "test"}),
            (PgVectorDB, {"embedder": Mock(), "db_url": "postgresql://localhost/test", "collection": "test"}),
            (MongoDBAtlas, {"embedder": Mock(), "connection_string": "mongodb://localhost:27017", "database_name": "test", "collection": "test"}),
        ],
    )
    async def test_vectordb_async_search(self, db_class, db_params, mock_embedding):
        db = db_class(**db_params)

        mock_result = VectorSearchResult(
            content="Test",
            metadata={},
            score=0.95,
        )

        if hasattr(db, "search_async"):
            with patch.object(db, "search_async", return_value=[mock_result]):
                results = await db.search_async("test query", top_k=5)
                assert isinstance(results, list)
        else:
            with patch.object(db, "search", return_value=[mock_result]):
                results = db.search("test query", top_k=5)
                assert isinstance(results, list)

    @pytest.mark.parametrize(
        "db_class,db_params",
        [
            (ChromaDB, {"embedder": Mock(), "collection": "test"}),
            (PineconeDB, {"api_key": "test", "embedder": Mock(), "collection": "test"}),
            (QdrantDB, {"embedder": Mock(), "collection": "test"}),
            (PgVectorDB, {"embedder": Mock(), "db_url": "postgresql://localhost/test", "collection": "test"}),
            (MongoDBAtlas, {"embedder": Mock(), "connection_string": "mongodb://localhost:27017", "database_name": "test", "collection": "test"}),
        ],
    )
    def test_vectordb_sync_add_documents(self, db_class, db_params, mock_embedding):
        db = db_class(**db_params)

        documents = [
            VectorDocument(
                id="doc_1",
                content="Test content",
                vector=mock_embedding,
                metadata={"source": "test"},
            )
        ]

        with patch.object(db, "add_documents", return_value=None):
            db.add_documents(documents)
            assert True

    @pytest.mark.parametrize(
        "db_class,db_params",
        [
            (ChromaDB, {"embedder": Mock(), "collection": "test"}),
            (PineconeDB, {"api_key": "test", "embedder": Mock(), "collection": "test"}),
            (QdrantDB, {"embedder": Mock(), "collection": "test"}),
            (PgVectorDB, {"embedder": Mock(), "db_url": "postgresql://localhost/test", "collection": "test"}),
            (MongoDBAtlas, {"embedder": Mock(), "connection_string": "mongodb://localhost:27017", "database_name": "test", "collection": "test"}),
        ],
    )
    def test_vectordb_sync_search(self, db_class, db_params, mock_embedding):
        db = db_class(**db_params)

        mock_result = VectorSearchResult(
            content="Test",
            metadata={},
            score=0.95,
        )

        with patch.object(db, "search", return_value=[mock_result]):
            results = db.search("test query", top_k=5)
            assert isinstance(results, list)


class TestVectorDBWithEmbedder:
    @pytest.fixture
    def mock_embedder(self):
        embedder = Mock()
        embedder.embed = AsyncMock(return_value=[0.1] * 100)
        embedder.embed_sync = Mock(return_value=[0.1] * 100)
        return embedder

    def test_vectordb_embedder_integration(self, mock_embedder):
        db = ChromaDB(embedder=mock_embedder, collection="test")
        assert db.embedder == mock_embedder

    @pytest.mark.asyncio
    async def test_vectordb_auto_embedding(self, mock_embedder):
        db = ChromaDB(embedder=mock_embedder, collection="test")

        assert db.embedder is not None



import asyncio
import warnings
from dataclasses import dataclass, field
from math import sqrt
from os import getenv
from typing import Any

from hypertic.utils.log import get_logger
from hypertic.vectordb.base import BaseVectorDB, VectorDocument, VectorSearchResult

warnings.filterwarnings("ignore", category=Warning, module="sqlalchemy")
warnings.filterwarnings("ignore", category=Warning, module="psycopg2")


def clean_error_message(error_msg: str) -> str:
    if "(Background on this error at:" in error_msg:
        return error_msg.split("(Background on this error at:")[0].strip()
    return error_msg


logger = get_logger(__name__)

try:
    from sqlalchemy import delete, func, select, text, update
    from sqlalchemy.dialects import postgresql
    from sqlalchemy.engine import Engine, create_engine
    from sqlalchemy.inspection import inspect
    from sqlalchemy.orm import Session, scoped_session, sessionmaker
    from sqlalchemy.schema import Column, MetaData, Table
    from sqlalchemy.types import DateTime, String, Text
except ImportError as err:
    raise ImportError("SQLAlchemy required for PgVectorDB. Install with: pip install sqlalchemy") from err

try:
    from pgvector.sqlalchemy import Vector
except ImportError as err:
    raise ImportError("pgvector required for PgVectorDB. Install with: pip install pgvector") from err


class Distance:
    COSINE = "cosine"
    L2 = "l2"
    INNER_PRODUCT = "inner_product"


class SearchType:
    VECTOR = "vector"
    KEYWORD = "keyword"
    HYBRID = "hybrid"


@dataclass
class HNSW:
    m: int = 16
    ef_construction: int = 64
    ef_search: int = 40
    name: str | None = None


@dataclass
class Ivfflat:
    lists: int = 100
    probes: int = 1
    name: str | None = None
    dynamic_lists: bool = True


@dataclass
class PgVectorDB(BaseVectorDB):
    collection: str
    db_url: str | None = None
    db_engine: Engine | None = None
    embedder: Any | None = None
    schema: str = "ai"
    vector_size: int = 1536
    distance_metric: str = Distance.COSINE
    search_type: str = SearchType.VECTOR
    vector_index: Ivfflat | HNSW = field(default_factory=HNSW)
    prefix_match: bool = False
    vector_score_weight: float = 0.5
    content_language: str = "english"

    table_name: str = field(init=False)

    _pool: Any | None = field(default=None, init=False)
    _sync_conn: Any | None = field(default=None, init=False)

    metadata: MetaData | None = field(default=None, init=False)
    _Session: Any | None = field(default=None, init=False)
    table: Any | None = field(default=None, init=False)

    def __post_init__(self):
        if not self.collection:
            raise ValueError("Collection name must be provided.")

        if self.db_url is None:
            self.db_url = getenv("PGVECTORDB_URL")

        if self.db_engine is None and self.db_url is None:
            raise ValueError("Either 'db_url' or 'db_engine' must be provided. Set PGVECTORDB_URL environment variable or pass db_url parameter.")

        BaseVectorDB.__init__(self, embedder=self.embedder)

        self.table_name = self.collection

        if self.db_engine is None:
            if self.db_url is None:
                raise ValueError("Must provide 'db_url' if 'db_engine' is None. Set PGVECTORDB_URL environment variable or pass db_url parameter.")
            try:
                self.db_engine = create_engine(self.db_url)
            except Exception as e:
                error_msg = str(e)
                if "psycopg2" in error_msg.lower() or "No module named 'psycopg2" in error_msg:
                    raise ImportError("Install with: pip install hypertic[pgvector] or pip install psycopg2-binary") from e
                logger.error(f"Failed to create engine from 'db_url': {e}", exc_info=True)
                raise

        self.metadata = MetaData(schema=self.schema)
        self._Session = scoped_session(sessionmaker(bind=self.db_engine))
        self.table = self._get_table()

    def _get_table(self):
        if self.metadata is None:
            raise RuntimeError("Metadata not initialized")
        return Table(
            self.table_name,
            self.metadata,
            Column("id", String, primary_key=True),
            Column("name", String),
            Column("meta_data", postgresql.JSONB, server_default=text("'{}'::jsonb")),
            Column("filters", postgresql.JSONB, server_default=text("'{}'::jsonb"), nullable=True),
            Column("content", Text),
            Column("embedding", Vector(self.vector_size)),
            Column("usage", postgresql.JSONB, server_default=text("'{}'::jsonb")),
            Column("created_at", DateTime(timezone=True), server_default=func.now()),
            Column("updated_at", DateTime(timezone=True), onupdate=func.now()),
            Column("content_hash", String),
            Column("content_id", String),
            extend_existing=True,
        )

    def table_exists(self) -> bool:
        try:
            if self.db_engine is None:
                return False
            inspector = inspect(self.db_engine)
            if inspector is None:
                return False
            return bool(inspector.has_table(self.table_name, schema=self.schema))
        except Exception as e:
            logger.error(f"Error checking if table exists: {e}", exc_info=True)
            return False

    async def _initialize_db(self) -> bool:
        try:
            if self.table_exists():
                return True

            if self._Session is None:
                raise RuntimeError("Session not initialized")
            with self._Session() as sess, sess.begin():
                sess.execute(text("CREATE EXTENSION IF NOT EXISTS vector;"))
                if self.schema is not None:
                    sess.execute(text(f"CREATE SCHEMA IF NOT EXISTS {self.schema};"))

            if self.table is None:
                raise RuntimeError("Table not initialized")
            self.table.create(self.db_engine)

            self._create_vector_index()

            self._create_gin_index()

            return True

        except Exception as e:
            logger.error(f"Error initializing PgVector database: {e}", exc_info=True)
            return False

    def _initialize_db_sync(self) -> bool:
        try:
            if self.table_exists():
                return True

            if self._Session is None:
                raise RuntimeError("Session not initialized")
            with self._Session() as sess, sess.begin():
                sess.execute(text("CREATE EXTENSION IF NOT EXISTS vector;"))
                if self.schema is not None:
                    sess.execute(text(f"CREATE SCHEMA IF NOT EXISTS {self.schema};"))

            if self.table is None:
                raise RuntimeError("Table not initialized")
            self.table.create(self.db_engine)

            self._create_vector_index()
            self._create_gin_index()

            return True

        except Exception as e:
            logger.error(f"Error initializing PgVector database: {e}", exc_info=True)
            return False

    def _create_vector_index(self, force_recreate: bool = False) -> None:
        if self.vector_index is None:
            return

        if self.vector_index.name is None:
            index_type = "ivfflat" if isinstance(self.vector_index, Ivfflat) else "hnsw"
            self.vector_index.name = f"{self.table_name}_{index_type}_index"

        index_distance = {
            Distance.L2: "vector_l2_ops",
            Distance.INNER_PRODUCT: "vector_ip_ops",
            Distance.COSINE: "vector_cosine_ops",
        }.get(self.distance_metric, "vector_cosine_ops")

        vector_index_exists = self._index_exists(self.vector_index.name)

        if vector_index_exists and not force_recreate:
            return

        if vector_index_exists and force_recreate:
            self._drop_index(self.vector_index.name)

        try:
            if self._Session is None:
                raise RuntimeError("Session not initialized")
            with self._Session() as sess, sess.begin():
                if isinstance(self.vector_index, Ivfflat):
                    self._create_ivfflat_index(sess, index_distance)
                elif isinstance(self.vector_index, HNSW):
                    self._create_hnsw_index(sess, index_distance)
        except Exception as e:
            logger.error(f"Error creating vector index '{self.vector_index.name}': {e}", exc_info=True)
            raise

    def _create_ivfflat_index(self, sess: Session, index_distance: str) -> None:
        if isinstance(self.vector_index, Ivfflat):
            num_lists = self.vector_index.lists
            if self.vector_index.dynamic_lists:
                total_records = self._count_documents_impl_sync()
                if total_records < 1000000:
                    num_lists = max(int(total_records / 1000), 1)
                else:
                    num_lists = max(int(sqrt(total_records)), 1)

            sess.execute(text("SET ivfflat.probes = :probes;"), {"probes": self.vector_index.probes})

            if self.table is None:
                raise RuntimeError("Table not initialized")
            index_name = self.vector_index.name or f"{self.table_name}_ivfflat_index"
            create_index_sql = text(
                f'CREATE INDEX "{index_name}" ON {self.table.fullname} USING ivfflat (embedding {index_distance}) WITH (lists = :num_lists);'
            )
            sess.execute(create_index_sql, {"num_lists": num_lists})
        else:
            raise ValueError("vector_index must be Ivfflat for IVFFlat index creation")

    def _create_hnsw_index(self, sess: Session, index_distance: str) -> None:
        if isinstance(self.vector_index, HNSW):
            if self.table is None:
                raise RuntimeError("Table not initialized")
            index_name = self.vector_index.name or f"{self.table_name}_hnsw_index"
            create_index_sql = text(
                f'CREATE INDEX "{index_name}" ON {self.table.fullname} '
                f"USING hnsw (embedding {index_distance}) "
                f"WITH (m = :m, ef_construction = :ef_construction);"
            )
            sess.execute(
                create_index_sql,
                {"m": self.vector_index.m, "ef_construction": self.vector_index.ef_construction},
            )
        else:
            raise ValueError("vector_index must be HNSW for HNSW index creation")

    def _create_gin_index(self, force_recreate: bool = False) -> None:
        gin_index_name = f"{self.table_name}_content_gin_index"

        gin_index_exists = self._index_exists(gin_index_name)

        if gin_index_exists and not force_recreate:
            return

        if gin_index_exists and force_recreate:
            self._drop_index(gin_index_name)

        try:
            if self._Session is None:
                raise RuntimeError("Session not initialized")
            with self._Session() as sess, sess.begin():
                if self.table is None:
                    raise RuntimeError("Table not initialized")
                create_gin_index_sql = text(
                    f"CREATE INDEX \"{gin_index_name}\" ON {self.table.fullname} USING GIN (to_tsvector('{self.content_language}', content));"
                )
                sess.execute(create_gin_index_sql)
        except Exception as e:
            logger.error(f"Error creating GIN index '{gin_index_name}': {e}", exc_info=True)
            raise

    def _index_exists(self, index_name: str) -> bool:
        if self.db_engine is None:
            return False
        inspector = inspect(self.db_engine)
        if inspector is None:
            return False
        if self.table is None:
            raise RuntimeError("Table not initialized")
        indexes = inspector.get_indexes(self.table.name, schema=self.schema)
        return any(idx["name"] == index_name for idx in indexes)

    def _drop_index(self, index_name: str) -> None:
        try:
            if self._Session is None:
                raise RuntimeError("Session not initialized")
            with self._Session() as sess, sess.begin():
                drop_index_sql = f'DROP INDEX IF EXISTS "{self.schema}"."{index_name}";'
                sess.execute(text(drop_index_sql))
        except Exception as e:
            logger.error(f"Error dropping index '{index_name}': {e}", exc_info=True)
            raise

    def _clean_content(self, content: str) -> str:
        return content.replace("\x00", "\ufffd")

    def _get_document_record(self, doc: VectorDocument, filters: dict[str, Any] | None = None, content_hash: str = "") -> dict[str, Any]:
        cleaned_content = self._clean_content(doc.content)
        record_id = doc.id or content_hash

        meta_data = doc.metadata or {}
        if filters:
            meta_data.update(filters)

        return {
            "id": record_id,
            "name": doc.id,
            "meta_data": meta_data,
            "filters": filters,
            "content": cleaned_content,
            "embedding": doc.vector,
            "usage": {},
            "content_hash": content_hash,
            "content_id": doc.id,
        }

    async def _add_documents_impl(self, documents: list[VectorDocument]) -> bool:
        try:
            if not await self.async_exists():
                await self._initialize_db()

            batch_records = []
            for doc in documents:
                try:
                    batch_records.append(self._get_document_record(doc))
                except Exception as e:
                    logger.error(f"Error processing document '{doc.id}': {e}", exc_info=True)

            if self._Session is None:
                raise RuntimeError("Session not initialized")
            if self.table is None:
                raise RuntimeError("Table not initialized")
            table_obj: Any = self.table
            with self._Session() as sess:
                insert_stmt = postgresql.insert(table_obj)
                sess.execute(insert_stmt, batch_records)
                sess.commit()

            return True

        except Exception as e:
            error_str = str(e).lower()
            if "uniqueviolation" in error_str or "duplicate key" in error_str:
                logger.info("Documents already exist in database (skipping duplicate insertion)")
                return True
            else:
                logger.error(
                    f"Error adding documents to PgVector: {clean_error_message(str(e))}",
                    exc_info=True,
                )
                return False

    def _add_documents_impl_sync(self, documents: list[VectorDocument]) -> bool:
        try:
            if not self.exists():
                self._initialize_db_sync()

            batch_records = []
            for doc in documents:
                try:
                    batch_records.append(self._get_document_record(doc))
                except Exception as e:
                    logger.error(f"Error processing document '{doc.id}': {e}", exc_info=True)

            if self._Session is None:
                raise RuntimeError("Session not initialized")
            if self.table is None:
                raise RuntimeError("Table not initialized")
            table_obj: Any = self.table
            with self._Session() as sess:
                insert_stmt = postgresql.insert(table_obj)
                sess.execute(insert_stmt, batch_records)
                sess.commit()

            return True

        except Exception as e:
            error_str = str(e).lower()
            if "uniqueviolation" in error_str or "duplicate key" in error_str:
                logger.info("Documents already exist in database (skipping duplicate insertion)")
                return True
            else:
                logger.error(
                    f"Error adding documents to PgVector: {clean_error_message(str(e))}",
                    exc_info=True,
                )
                return False

    async def _search_impl(self, query_vector: list[float], top_k: int = 5, filters: dict[str, Any] | None = None) -> list[VectorSearchResult]:
        return await asyncio.to_thread(self._search_impl_sync, query_vector, top_k, filters)

    def _search_impl_sync(self, query_vector: list[float], top_k: int, filters: dict[str, Any] | None = None) -> list[VectorSearchResult]:
        try:
            if not self.exists():
                return []

            if self.table is None:
                raise RuntimeError("Table not initialized")
            columns = [
                self.table.c.id,
                self.table.c.name,
                self.table.c.meta_data,
                self.table.c.content,
                self.table.c.embedding,
            ]

            stmt = select(*columns)

            if filters is not None:
                if self.table is None:
                    raise RuntimeError("Table not initialized")
                stmt = stmt.where(self.table.c.meta_data.contains(filters))

            if self.distance_metric == Distance.L2:
                stmt = stmt.order_by(self.table.c.embedding.l2_distance(query_vector))
            elif self.distance_metric == Distance.COSINE:
                stmt = stmt.order_by(self.table.c.embedding.cosine_distance(query_vector))
            elif self.distance_metric == Distance.INNER_PRODUCT:
                stmt = stmt.order_by(self.table.c.embedding.max_inner_product(query_vector))
            else:
                logger.warning(f"Unknown distance metric: {self.distance_metric}")
                return []

            stmt = stmt.limit(top_k)

            try:
                if self._Session is None:
                    raise RuntimeError("Session not initialized")
                with self._Session() as sess, sess.begin():
                    if self.vector_index is not None:
                        if isinstance(self.vector_index, Ivfflat):
                            sess.execute(text(f"SET LOCAL ivfflat.probes = {self.vector_index.probes}"))
                        elif isinstance(self.vector_index, HNSW):
                            sess.execute(text(f"SET LOCAL hnsw.ef_search = {self.vector_index.ef_search}"))
                    results = sess.execute(stmt).fetchall()
            except Exception as e:
                logger.error(f"Error performing vector search: {e}", exc_info=True)
                return []

            search_results = []
            for result in results:
                similarity: float
                if self.distance_metric == Distance.COSINE:
                    import numpy as np

                    embedding_array = np.array(result.embedding)
                    query_array = np.array(query_vector)

                    dot_product = np.dot(embedding_array, query_array)
                    norm_a = np.linalg.norm(embedding_array)
                    norm_b = np.linalg.norm(query_array)

                    if norm_a == 0 or norm_b == 0:
                        similarity = 0.0
                    else:
                        similarity = float(dot_product / (norm_a * norm_b))

                elif self.distance_metric == Distance.L2:
                    import numpy as np

                    embedding_array = np.array(result.embedding)
                    query_array = np.array(query_vector)

                    l2_distance = np.linalg.norm(embedding_array - query_array)
                    similarity = float(1 / (1 + l2_distance))

                else:
                    import numpy as np

                    embedding_array = np.array(result.embedding)
                    query_array = np.array(query_vector)

                    inner_product = np.dot(embedding_array, query_array)
                    similarity = float(max(0, min(1, (inner_product + 1) / 2)))

                search_results.append(
                    VectorSearchResult(
                        content=result.content,
                        score=similarity,
                        metadata=result.meta_data if result.meta_data else {},
                    )
                )

            return search_results

        except Exception as e:
            logger.error(f"Error searching PgVector: {clean_error_message(str(e))}", exc_info=True)
            return []

    async def _get_documents_impl(
        self,
        ids: list[str] | None = None,
        where: dict[str, Any] | None = None,
        limit: int | None = None,
        include: list[str] | None = None,
    ) -> dict[str, Any]:
        try:
            if not await self.async_exists():
                return {"ids": [], "contents": [], "metadatas": []}

            if self.table is None:
                raise RuntimeError("Table not initialized")
            columns = [self.table.c.id, self.table.c.content, self.table.c.meta_data]
            stmt = select(*columns)

            if ids:
                if self.table is None:
                    raise RuntimeError("Table not initialized")
                stmt = stmt.where(self.table.c.id.in_(ids))
            if limit:
                stmt = stmt.limit(limit)

            if self._Session is None:
                raise RuntimeError("Session not initialized")
            with self._Session() as sess, sess.begin():
                results = sess.execute(stmt).fetchall()

                result_ids = []
                contents = []
                metadatas = []

                for result in results:
                    result_ids.append(result.id)
                    contents.append(result.content)
                    metadatas.append(result.meta_data if result.meta_data else {})

                return {"ids": result_ids, "contents": contents, "metadatas": metadatas}

        except Exception as e:
            logger.error(
                f"Error getting documents from PgVector: {clean_error_message(str(e))}",
                exc_info=True,
            )
            return {"ids": [], "contents": [], "metadatas": []}

    def _get_documents_impl_sync(
        self,
        ids: list[str] | None = None,
        where: dict[str, Any] | None = None,
        limit: int | None = None,
        include: list[str] | None = None,
    ) -> dict[str, Any]:
        try:
            if not self.exists():
                return {"ids": [], "contents": [], "metadatas": []}

            if self.table is None:
                raise RuntimeError("Table not initialized")
            columns = [self.table.c.id, self.table.c.content, self.table.c.meta_data]
            stmt = select(*columns)

            if ids:
                if self.table is None:
                    raise RuntimeError("Table not initialized")
                stmt = stmt.where(self.table.c.id.in_(ids))
            if limit:
                stmt = stmt.limit(limit)

            if self._Session is None:
                raise RuntimeError("Session not initialized")
            with self._Session() as sess, sess.begin():
                results = sess.execute(stmt).fetchall()

                result_ids = []
                contents = []
                metadatas = []

                for result in results:
                    result_ids.append(result.id)
                    contents.append(result.content)
                    metadatas.append(result.meta_data if result.meta_data else {})

                return {"ids": result_ids, "contents": contents, "metadatas": metadatas}

        except Exception as e:
            logger.error(
                f"Error getting documents from PgVector: {clean_error_message(str(e))}",
                exc_info=True,
            )
            return {"ids": [], "contents": [], "metadatas": []}

    async def _count_documents_impl(self) -> int:
        try:
            if not await self.async_exists():
                return 0

            if self._Session is None:
                raise RuntimeError("Session not initialized")
            with self._Session() as sess, sess.begin():
                if self.table is None:
                    raise RuntimeError("Table not initialized")
                stmt = select(func.count(self.table.c.id))
                result = sess.execute(stmt).scalar()
                return int(result) if result is not None else 0

        except Exception as e:
            logger.error(f"Error counting documents in PgVector: {e}", exc_info=True)
            return 0

    def _count_documents_impl_sync(self) -> int:
        try:
            if not self.exists():
                return 0

            if self._Session is None:
                raise RuntimeError("Session not initialized")
            with self._Session() as sess, sess.begin():
                if self.table is None:
                    raise RuntimeError("Table not initialized")
                stmt = select(func.count(self.table.c.id))
                result = sess.execute(stmt).scalar()
                return int(result) if result is not None else 0

        except Exception as e:
            logger.error(f"Error counting documents in PgVector: {e}", exc_info=True)
            return 0

    async def async_exists(self) -> bool:
        return await asyncio.to_thread(self.table_exists)

    def exists(self) -> bool:
        return self.table_exists()

    def keyword_search(self, query: str, top_k: int) -> list[VectorSearchResult]:
        try:
            if not self.exists():
                return []

            if self.table is None:
                raise RuntimeError("Table not initialized")

            search_query = func.to_tsquery(self.content_language, query)
            rank_func = func.ts_rank(func.to_tsvector(self.content_language, self.table.c.content), search_query)

            stmt = (
                select(
                    self.table.c.id,
                    self.table.c.content,
                    self.table.c.meta_data,
                    rank_func.label("rank"),
                )
                .where(func.to_tsvector(self.content_language, self.table.c.content).match(search_query))
                .order_by(rank_func.desc())
                .limit(top_k)
            )

            if self._Session is None:
                raise RuntimeError("Session not initialized")
            with self._Session() as sess, sess.begin():
                results = sess.execute(stmt).fetchall()

            search_results = []
            for result in results:
                search_results.append(
                    VectorSearchResult(
                        content=result.content or "",
                        score=float(result.rank) if result.rank is not None else 0.0,
                        metadata=result.meta_data if result.meta_data else {},
                    )
                )

            return search_results

        except Exception as e:
            logger.error(f"Error performing keyword search: {e}", exc_info=True)
            return []

    async def _search_text_impl(self, query: str, top_k: int) -> list[VectorSearchResult]:
        return await asyncio.to_thread(self.keyword_search, query, top_k)

    def _search_text_impl_sync(self, query: str, top_k: int) -> list[VectorSearchResult]:
        return self.keyword_search(query, top_k)

    async def _delete_impl(self, ids: list[str] | None = None, where: dict[str, Any] | None = None) -> bool:
        return await asyncio.to_thread(self._delete_impl_sync, ids, where)

    def _delete_impl_sync(self, ids: list[str] | None = None, where: dict[str, Any] | None = None) -> bool:
        try:
            if not self.exists():
                return False

            if self._Session is None:
                raise RuntimeError("Session not initialized")
            with self._Session() as sess, sess.begin():
                if ids is not None:
                    if self.table is None:
                        raise RuntimeError("Table not initialized")
                    stmt = delete(self.table).where(self.table.c.id.in_(ids))
                elif where is not None:
                    if self.table is None:
                        raise RuntimeError("Table not initialized")
                    stmt = delete(self.table).where(self.table.c.meta_data.contains(where))
                else:
                    return False

                result = sess.execute(stmt)
                return bool(result.rowcount and result.rowcount > 0)

        except Exception as e:
            logger.error(
                f"Error deleting documents from PgVector: {clean_error_message(str(e))}",
                exc_info=True,
            )
            return False

    async def _update_impl(
        self,
        ids: list[str],
        metadatas: list[dict[str, Any]] | None = None,
        documents: list[str] | None = None,
        embeddings: list[list[float]] | None = None,
    ) -> bool:
        return await asyncio.to_thread(self._update_impl_sync, ids, metadatas, documents, embeddings)

    def _update_impl_sync(
        self,
        ids: list[str],
        metadatas: list[dict[str, Any]] | None = None,
        documents: list[str] | None = None,
        embeddings: list[list[float]] | None = None,
    ) -> bool:
        try:
            if not self.exists():
                return False

            if self._Session is None:
                raise RuntimeError("Session not initialized")
            with self._Session() as sess, sess.begin():
                for i, doc_id in enumerate(ids):
                    update_data = {}

                    if metadatas and i < len(metadatas):
                        update_data["meta_data"] = metadatas[i]

                    if documents and i < len(documents):
                        content_value: Any = documents[i]
                        update_data["content"] = content_value
                        if self.embedder and embeddings is None:
                            new_embedding = self.embedder.embed_sync(documents[i])
                            update_data["embedding"] = new_embedding

                    if embeddings and i < len(embeddings):
                        embedding_value: Any = embeddings[i]
                        update_data["embedding"] = embedding_value

                    if update_data:
                        if self.table is None:
                            raise RuntimeError("Table not initialized")
                        stmt = update(self.table).where(self.table.c.id == doc_id).values(**update_data)
                        sess.execute(stmt)

                return True

        except Exception as e:
            logger.error(
                f"Error updating documents in PgVector: {clean_error_message(str(e))}",
                exc_info=True,
            )
            return False

    async def _upsert_impl(
        self,
        ids: list[str],
        metadatas: list[dict[str, Any]] | None = None,
        documents: list[str] | None = None,
        embeddings: list[list[float]] | None = None,
    ) -> bool:
        return await asyncio.to_thread(self._upsert_impl_sync, ids, metadatas, documents, embeddings)

    def _upsert_impl_sync(
        self,
        ids: list[str],
        metadatas: list[dict[str, Any]] | None = None,
        documents: list[str] | None = None,
        embeddings: list[list[float]] | None = None,
    ) -> bool:
        try:
            if not self.exists():
                self._initialize_db_sync()

            if self._Session is None:
                raise RuntimeError("Session not initialized")
            with self._Session() as sess, sess.begin():
                for i, doc_id in enumerate(ids):
                    content = documents[i] if documents and i < len(documents) else ""
                    metadata = metadatas[i] if metadatas and i < len(metadatas) else {}
                    embedding = embeddings[i] if embeddings and i < len(embeddings) else None

                    if embedding is None and content and self.embedder:
                        embedding = self.embedder.embed_sync(content)

                    record_data = {
                        "id": doc_id,
                        "name": doc_id,
                        "meta_data": metadata,
                        "content": content,
                        "embedding": embedding,
                        "usage": {},
                        "content_hash": "",
                        "content_id": doc_id,
                    }

                    from sqlalchemy.dialects import postgresql

                    if self.table is None:
                        raise RuntimeError("Table not initialized")
                    insert_stmt = postgresql.insert(self.table).values(record_data)
                    upsert_stmt = insert_stmt.on_conflict_do_update(
                        index_elements=["id"],
                        set_={
                            "name": insert_stmt.excluded.name,
                            "meta_data": insert_stmt.excluded.meta_data,
                            "content": insert_stmt.excluded.content,
                            "embedding": insert_stmt.excluded.embedding,
                            "usage": insert_stmt.excluded.usage,
                            "content_hash": insert_stmt.excluded.content_hash,
                            "content_id": insert_stmt.excluded.content_id,
                        },
                    )
                    sess.execute(upsert_stmt)

                return True

        except Exception as e:
            logger.error(
                f"Error upserting documents in PgVector: {clean_error_message(str(e))}",
                exc_info=True,
            )
            return False



import asyncio
import contextlib
from dataclasses import dataclass, field
from os import getenv
from typing import Any

from pymongo import AsyncMongoClient, MongoClient
from pymongo.collection import Collection
from pymongo.database import Database
from pymongo.operations import SearchIndexModel

from hypertic.utils.log import get_logger
from hypertic.vectordb.base import BaseVectorDB, VectorDocument, VectorSearchResult

logger = get_logger(__name__)


@dataclass
class MongoDBAtlas(BaseVectorDB):
    collection: str
    embedder: Any | None = None
    connection_string: str | None = None
    database_name: str = "vectordb"
    vector_field: str = "embedding"

    collection_name: str = field(init=False)

    _client: MongoClient[Any] | None = field(default=None, init=False)
    _database: Database[Any] | None = field(default=None, init=False)
    _collection: Collection[Any] | None = field(default=None, init=False)
    _async_client: AsyncMongoClient[Any] | None = field(default=None, init=False)
    _async_database: Any = field(default=None, init=False)
    _async_collection: Collection[Any] | None = field(default=None, init=False)
    _initialized: bool = field(default=False, init=False)

    def __post_init__(self):
        BaseVectorDB.__init__(self, embedder=self.embedder)

        self.collection_name = self.collection

        self.connection_string = self.connection_string or getenv("MONGODB_VECTORDB_URL")

        if not self.connection_string:
            raise ValueError(
                "MongoDB connection string is required. Pass connection_string parameter or set MONGODB_VECTORDB_URL environment variable."
            )

    async def _maybe_await_async(self, result: Any) -> Any:
        if hasattr(result, "__await__"):
            return await result
        return result

    def _get_client(self) -> MongoClient[Any]:
        if self._client is None:
            self._client = MongoClient(self.connection_string)
        return self._client

    def _get_database(self) -> Database[Any]:
        if self._database is None:
            client = self._get_client()
            self._database = client[self.database_name]
        return self._database

    def _get_collection(self) -> Collection[Any]:
        if self._collection is None:
            db = self._get_database()
            self._collection = db[self.collection_name]
        return self._collection

    async def _get_async_client(self) -> AsyncMongoClient[Any]:
        if self._async_client is None:
            self._async_client = AsyncMongoClient(
                self.connection_string,
                maxPoolSize=100,
                retryWrites=True,
                serverSelectionTimeoutMS=5000,
            )
            await self._async_client.admin.command("ping")
            database_result = self._async_client[self.database_name]
            self._async_database = database_result
        return self._async_client

    async def _get_async_collection(self) -> Collection[Any]:
        if self._async_collection is None:
            await self._get_async_client()
            if self._async_database is None:
                raise RuntimeError("Async database not initialized")
            collection_result = self._async_database[self.collection_name]
            self._async_collection = collection_result
        return self._async_collection

    async def _initialize_db(self) -> bool:
        try:
            await self._get_async_client()
            collection = await self._get_async_collection()

            try:
                await self._maybe_await_async(collection.find_one())
            except Exception:
                await self._maybe_await_async(collection.insert_one({"_id": "__init_placeholder__"}))
                await self._maybe_await_async(collection.delete_one({"_id": "__init_placeholder__"}))
                await asyncio.sleep(0.1)

            # Try to create vector search index, but don't fail initialization if it fails
            try:
                await self._create_vector_search_index()
            except Exception as e:
                logger.debug(f"Vector search index creation failed (non-critical): {e}")

            self._initialized = True
            return True

        except Exception as e:
            logger.error(f"Error initializing MongoDB Atlas database: {e}", exc_info=True)
            return False

    def _initialize_db_sync(self) -> bool:
        try:
            client = self._get_client()
            client.admin.command("ping")

            self._get_database()
            collection = self._get_collection()

            try:
                collection.find_one()
            except Exception:
                collection.insert_one({"_id": "__placeholder__"})
                collection.delete_one({"_id": "__placeholder__"})

            self._create_vector_search_index_sync()

            self._initialized = True
            return True

        except Exception as e:
            logger.error(f"Error initializing MongoDB Atlas database: {e}", exc_info=True)
            return False

    async def _create_vector_search_index(self):
        try:
            collection = await self._get_async_collection()

            try:
                await self._maybe_await_async(collection.find_one())
            except Exception as e:
                logger.debug(f"Collection not ready for index creation yet: {e}")
                return

            try:
                cursor_result = collection.list_search_indexes()
                cursor = await self._maybe_await_async(cursor_result)
                if cursor is None:
                    existing_indexes = []
                else:
                    to_list_result = cursor.to_list(None)
                    existing_indexes = await self._maybe_await_async(to_list_result)
                    if existing_indexes is None:
                        existing_indexes = []
            except Exception as e:
                logger.debug(f"Error listing search indexes (may not exist yet): {e}")
                existing_indexes = []

            index_exists = any(idx.get("name") == f"{self.collection_name}_vector_index" for idx in existing_indexes)

            if not index_exists:
                search_index_model = SearchIndexModel(
                    definition={
                        "fields": [
                            {
                                "type": "vector",
                                "path": self.vector_field,
                                "numDimensions": 1536,
                                "similarity": "cosine",
                            }
                        ]
                    },
                    name=f"{self.collection_name}_vector_index",
                    type="vectorSearch",
                )

                await self._maybe_await_async(collection.create_search_index(model=search_index_model))

        except Exception as e:
            error_msg = str(e).lower()
            if "does not exist" in error_msg or "namespace not found" in error_msg:
                logger.debug(f"Vector search index creation deferred (collection not ready): {e}")
            else:
                logger.error(f"Error creating vector search index: {e}", exc_info=True)

    def _create_vector_search_index_sync(self):
        try:
            collection = self._get_collection()

            try:
                collection.find_one()
            except Exception as e:
                logger.debug(f"Collection not ready for index creation yet: {e}")
                return

            try:
                existing_indexes = list(collection.list_search_indexes())
            except Exception:
                existing_indexes = []

            index_exists = any(idx.get("name") == f"{self.collection_name}_vector_index" for idx in existing_indexes)

            if not index_exists:
                search_index_model = SearchIndexModel(
                    definition={
                        "fields": [
                            {
                                "type": "vector",
                                "path": self.vector_field,
                                "numDimensions": 1536,
                                "similarity": "cosine",
                            }
                        ]
                    },
                    name=f"{self.collection_name}_vector_index",
                    type="vectorSearch",
                )

                collection.create_search_index(model=search_index_model)

        except Exception as e:
            error_msg = str(e).lower()
            if "does not exist" in error_msg or "namespace not found" in error_msg:
                logger.debug(f"Vector search index creation deferred (collection not ready): {e}")
            else:
                logger.error(f"Error creating vector search index: {e}", exc_info=True)

    async def _add_documents_impl(self, documents: list[VectorDocument]) -> bool:
        try:
            if not self._initialized:
                await self._initialize_db()

            collection = await self._get_async_collection()

            for doc in documents:
                if doc.vector is None:
                    logger.warning(f"Document {doc.id} has no vector, skipping")
                    continue

                doc_dict = {
                    "_id": doc.id,
                    self.vector_field: doc.vector,
                    "content": doc.content,
                    "metadata": doc.metadata or {},
                }

                await self._maybe_await_async(collection.replace_one({"_id": doc.id}, doc_dict, upsert=True))

            await self._create_vector_search_index()

            return True

        except Exception as e:
            logger.error(f"Error adding documents to MongoDB Atlas: {e}", exc_info=True)
            return False

    def _add_documents_impl_sync(self, documents: list[VectorDocument]) -> bool:
        try:
            if not self._initialized:
                self._initialize_db_sync()

            collection = self._get_collection()

            for doc in documents:
                if doc.vector is None:
                    logger.warning(f"Document {doc.id} has no vector, skipping")
                    continue

                doc_dict = {
                    "_id": doc.id,
                    self.vector_field: doc.vector,
                    "content": doc.content,
                    "metadata": doc.metadata or {},
                }

                collection.replace_one({"_id": doc.id}, doc_dict, upsert=True)

            self._create_vector_search_index_sync()

            return True

        except Exception as e:
            logger.error(f"Error adding documents to MongoDB Atlas: {e}", exc_info=True)
            return False

    async def _search_impl(self, query_vector: list[float], top_k: int) -> list[VectorSearchResult]:
        try:
            if not self._initialized:
                await self._initialize_db()

            collection = await self._get_async_collection()

            pipeline = [
                {
                    "$vectorSearch": {
                        "index": f"{self.collection_name}_vector_index",
                        "path": self.vector_field,
                        "queryVector": query_vector,
                        "numCandidates": top_k * 10,
                        "limit": top_k,
                    }
                },
                {
                    "$project": {
                        "_id": 1,
                        "content": 1,
                        "metadata": 1,
                        "score": {"$meta": "vectorSearchScore"},
                    }
                },
            ]

            aggregate_result = collection.aggregate(pipeline)
            cursor = await self._maybe_await_async(aggregate_result)
            to_list_result = cursor.to_list(length=None)
            results = await self._maybe_await_async(to_list_result)

            search_results = []
            for result in results:
                search_results.append(
                    VectorSearchResult(
                        content=result.get("content", ""),
                        metadata=result.get("metadata", {}),
                        score=result.get("score", 0.0),
                    )
                )

            return search_results

        except Exception as e:
            logger.error(f"Error searching MongoDB Atlas: {e}", exc_info=True)
            return []

    def _search_impl_sync(self, query_vector: list[float], top_k: int) -> list[VectorSearchResult]:
        try:
            if not self._initialized:
                self._initialize_db_sync()

            collection = self._get_collection()

            pipeline = [
                {
                    "$vectorSearch": {
                        "index": f"{self.collection_name}_vector_index",
                        "path": self.vector_field,
                        "queryVector": query_vector,
                        "numCandidates": top_k * 10,
                        "limit": top_k,
                    }
                },
                {
                    "$project": {
                        "_id": 1,
                        "content": 1,
                        "metadata": 1,
                        "score": {"$meta": "vectorSearchScore"},
                    }
                },
            ]

            results = list(collection.aggregate(pipeline))

            search_results = []
            for result in results:
                search_results.append(
                    VectorSearchResult(
                        content=result.get("content", ""),
                        metadata=result.get("metadata", {}),
                        score=result.get("score", 0.0),
                    )
                )

            return search_results

        except Exception as e:
            logger.error(f"Error searching MongoDB Atlas: {e}", exc_info=True)
            return []

    async def _search_text_impl(self, query: str, top_k: int) -> list[VectorSearchResult]:
        if self.embedder:
            query_vector = await self.embedder.embed(query)
            return await self._search_impl(query_vector, top_k)
        else:
            return []

    def _search_text_impl_sync(self, query: str, top_k: int) -> list[VectorSearchResult]:
        if self.embedder:
            query_vector = self.embedder.embed_sync(query)
            return self._search_impl_sync(query_vector, top_k)
        else:
            return []

    async def _delete_impl(self, ids: list[str] | None = None, where: dict[str, Any] | None = None) -> bool:
        try:
            if not self._initialized:
                await self._initialize_db()

            collection = self._get_collection()

            if ids:
                result = await asyncio.to_thread(collection.delete_many, {"_id": {"$in": ids}})
            elif where:
                result = await asyncio.to_thread(collection.delete_many, where)
            else:
                logger.warning("No IDs or filter provided for deletion")
                return False

            return bool(result.deleted_count > 0)

        except Exception as e:
            logger.error(f"Error deleting documents from MongoDB Atlas: {e}", exc_info=True)
            return False

    def _delete_impl_sync(self, ids: list[str] | None = None, where: dict[str, Any] | None = None) -> bool:
        try:
            if not self._initialized:
                self._initialize_db_sync()

            collection = self._get_collection()

            if ids:
                result = collection.delete_many({"_id": {"$in": ids}})
            elif where:
                result = collection.delete_many(where)
            else:
                logger.warning("No IDs or filter provided for deletion")
                return False

            return bool(result.deleted_count > 0)

        except Exception as e:
            logger.error(f"Error deleting documents from MongoDB Atlas: {e}", exc_info=True)
            return False

    async def _update_impl(
        self,
        ids: list[str],
        metadatas: list[dict[str, Any]] | None = None,
        documents: list[str] | None = None,
        embeddings: list[list[float]] | None = None,
    ) -> bool:
        try:
            if not self._initialized:
                await self._initialize_db()

            collection = self._get_collection()

            for i, doc_id in enumerate(ids):
                update_doc: dict[str, Any] = {}

                if metadatas and i < len(metadatas):
                    update_doc["metadata"] = metadatas[i]

                if documents and i < len(documents):
                    update_doc["content"] = documents[i]

                if embeddings and i < len(embeddings):
                    update_doc[self.vector_field] = embeddings[i]

                if update_doc:
                    await asyncio.to_thread(collection.update_one, {"_id": doc_id}, {"$set": update_doc})

            return True

        except Exception as e:
            logger.error(f"Error updating documents in MongoDB Atlas: {e}", exc_info=True)
            return False

    def _update_impl_sync(
        self,
        ids: list[str],
        metadatas: list[dict[str, Any]] | None = None,
        documents: list[str] | None = None,
        embeddings: list[list[float]] | None = None,
    ) -> bool:
        try:
            if not self._initialized:
                self._initialize_db_sync()

            collection = self._get_collection()

            for i, doc_id in enumerate(ids):
                update_doc: dict[str, Any] = {}

                if metadatas and i < len(metadatas):
                    update_doc["metadata"] = metadatas[i]

                if documents and i < len(documents):
                    update_doc["content"] = documents[i]

                if embeddings and i < len(embeddings):
                    update_doc[self.vector_field] = embeddings[i]

                if update_doc:
                    collection.update_one({"_id": doc_id}, {"$set": update_doc})

            return True

        except Exception as e:
            logger.error(f"Error updating documents in MongoDB Atlas: {e}", exc_info=True)
            return False

    async def _upsert_impl(
        self,
        ids: list[str],
        metadatas: list[dict[str, Any]] | None = None,
        documents: list[str] | None = None,
        embeddings: list[list[float]] | None = None,
    ) -> bool:
        try:
            if not self._initialized:
                await self._initialize_db()

            collection = self._get_collection()

            for i, doc_id in enumerate(ids):
                doc_dict: dict[str, Any] = {"_id": doc_id}

                if metadatas and i < len(metadatas):
                    doc_dict["metadata"] = metadatas[i]

                content = documents[i] if documents and i < len(documents) else None
                if content:
                    doc_dict["content"] = content
                    if not embeddings or i >= len(embeddings):
                        if self.embedder:
                            doc_dict[self.vector_field] = await self.embedder.embed(content)
                    else:
                        doc_dict[self.vector_field] = embeddings[i]
                else:
                    if embeddings and i < len(embeddings):
                        doc_dict[self.vector_field] = embeddings[i]

                await asyncio.to_thread(collection.replace_one, {"_id": doc_id}, doc_dict, upsert=True)

            return True

        except Exception as e:
            logger.error(f"Error upserting documents in MongoDB Atlas: {e}", exc_info=True)
            return False

    def _upsert_impl_sync(
        self,
        ids: list[str],
        metadatas: list[dict[str, Any]] | None = None,
        documents: list[str] | None = None,
        embeddings: list[list[float]] | None = None,
    ) -> bool:
        try:
            if not self._initialized:
                self._initialize_db_sync()

            collection = self._get_collection()

            for i, doc_id in enumerate(ids):
                doc_dict: dict[str, Any] = {"_id": doc_id}

                if metadatas and i < len(metadatas):
                    doc_dict["metadata"] = metadatas[i]

                content = documents[i] if documents and i < len(documents) else None
                if content:
                    doc_dict["content"] = content
                    if not embeddings or i >= len(embeddings):
                        if self.embedder:
                            doc_dict[self.vector_field] = self.embedder.embed_sync(content)
                    else:
                        doc_dict[self.vector_field] = embeddings[i]
                else:
                    if embeddings and i < len(embeddings):
                        doc_dict[self.vector_field] = embeddings[i]

                collection.replace_one({"_id": doc_id}, doc_dict, upsert=True)

            return True

        except Exception as e:
            logger.error(f"Error upserting documents in MongoDB Atlas: {e}", exc_info=True)
            return False

    async def _get_documents_impl(
        self,
        ids: list[str] | None = None,
        where: dict[str, Any] | None = None,
        limit: int | None = None,
        include: list[str] | None = None,
    ) -> dict[str, Any]:
        try:
            if not self._initialized:
                await self._initialize_db()

            collection = await self._get_async_collection()

            query = {}
            if ids:
                query["_id"] = {"$in": ids}
            elif where:
                for key, value in where.items():
                    query[f"metadata.{key}"] = value

            projection = {"_id": 1, "content": 1, "metadata": 1}
            if include:
                actual_fields = []
                for field in include:
                    if field == "documents":
                        actual_fields.append("content")
                    elif field == "embeddings":
                        actual_fields.append(self.vector_field)
                    elif field == "ids":
                        actual_fields.append("_id")
                    else:
                        actual_fields.append(field)
                projection = dict.fromkeys(actual_fields, 1)

            cursor = collection.find(query, projection)
            if limit:
                cursor = cursor.limit(limit)

            to_list_result = cursor.to_list(length=None)
            results = await self._maybe_await_async(to_list_result)

            result_ids = [doc.get("_id", "") for doc in results]
            contents = [doc.get("content", "") for doc in results]
            metadatas = [doc.get("metadata", {}) for doc in results]
            vectors = [doc.get(self.vector_field, []) for doc in results] if include and "embeddings" in include else []

            return_dict = {"ids": result_ids}
            if include:
                if "documents" in include:
                    return_dict["documents"] = contents
                if "contents" in include:
                    return_dict["contents"] = contents
                if "metadatas" in include:
                    return_dict["metadatas"] = metadatas
                if "embeddings" in include:
                    return_dict["embeddings"] = vectors
            else:
                return_dict.update({"documents": contents, "contents": contents, "metadatas": metadatas})

            return return_dict

        except Exception as e:
            logger.error(f"Error getting documents from MongoDB Atlas: {e}", exc_info=True)
            return {"ids": [], "contents": [], "metadatas": []}

    def _get_documents_impl_sync(
        self,
        ids: list[str] | None = None,
        where: dict[str, Any] | None = None,
        limit: int | None = None,
        include: list[str] | None = None,
    ) -> dict[str, Any]:
        try:
            if not self._initialized:
                self._initialize_db_sync()

            collection = self._get_collection()

            query = {}
            if ids:
                query["_id"] = {"$in": ids}
            elif where:
                for key, value in where.items():
                    query[f"metadata.{key}"] = value

            projection = {"_id": 1, "content": 1, "metadata": 1}
            if include:
                actual_fields = []
                for field in include:
                    if field == "documents":
                        actual_fields.append("content")
                    elif field == "embeddings":
                        actual_fields.append(self.vector_field)
                    elif field == "ids":
                        actual_fields.append("_id")
                    else:
                        actual_fields.append(field)
                projection = dict.fromkeys(actual_fields, 1)

            cursor = collection.find(query, projection)
            if limit:
                cursor = cursor.limit(limit)

            results = list(cursor)

            result_ids = [doc.get("_id", "") for doc in results]
            contents = [doc.get("content", "") for doc in results]
            metadatas = [doc.get("metadata", {}) for doc in results]
            vectors = [doc.get(self.vector_field, []) for doc in results] if include and "embeddings" in include else []

            return_dict = {"ids": result_ids}
            if include:
                if "documents" in include:
                    return_dict["documents"] = contents
                if "contents" in include:
                    return_dict["contents"] = contents
                if "metadatas" in include:
                    return_dict["metadatas"] = metadatas
                if "embeddings" in include:
                    return_dict["embeddings"] = vectors
            else:
                return_dict.update({"documents": contents, "contents": contents, "metadatas": metadatas})

            return return_dict

        except Exception as e:
            logger.error(f"Error getting documents from MongoDB Atlas: {e}", exc_info=True)
            return {"ids": [], "contents": [], "metadatas": []}

    async def _count_documents_impl(self) -> int:
        try:
            if not self._initialized:
                await self._initialize_db()

            collection = await self._get_async_collection()
            count_result = collection.count_documents({})
            count = await self._maybe_await_async(count_result)
            return int(count) if count is not None else 0

        except Exception as e:
            logger.error(f"Error counting documents in MongoDB Atlas: {e}", exc_info=True)
            return 0

    def _count_documents_impl_sync(self) -> int:
        try:
            if not self._initialized:
                self._initialize_db_sync()

            collection = self._get_collection()
            count = collection.count_documents({})
            return int(count)

        except Exception as e:
            logger.error(f"Error counting documents in MongoDB Atlas: {e}", exc_info=True)
            return 0

    def exists(self) -> bool:
        try:
            if not self._initialized:
                with contextlib.suppress(Exception):
                    self._initialize_db_sync()

            collection = self._get_collection()
            collection.database.list_collection_names()
            collections = collection.database.list_collection_names()
            return self.collection_name in collections

        except Exception:
            return False



import asyncio
import dataclasses
import json
import uuid
from collections.abc import AsyncIterator, Iterator
from dataclasses import dataclass, field
from typing import Any, Optional, cast, get_origin, get_type_hints

from pydantic import BaseModel, create_model

from hypertic.models.base import LLMResponse
from hypertic.models.events import (
    MetadataEvent,
    StreamEvent,
)
from hypertic.tools.base import BaseToolkit, _ToolManager
from hypertic.tools.mcp.client import ExecutableTool
from hypertic.utils.exceptions import (
    ConfigurationError,
    GuardrailViolationError,
    MaxStepsError,
    RetrieverError,
    SchemaConversionError,
    ToolExecutionError,
    ToolNotFoundError,
)
from hypertic.utils.files import FileProcessor
from hypertic.utils.log import get_logger

logger = get_logger(__name__)


def _convert_to_pydantic(output_type: type[Any]) -> type[BaseModel]:
    if issubclass(output_type, BaseModel):
        return output_type

    if dataclasses.is_dataclass(output_type):
        hints = get_type_hints(output_type)
        field_definitions = {}
        for field in dataclasses.fields(output_type):
            field_type = hints.get(field.name, field.type)
            if field.default is not dataclasses.MISSING:
                field_definitions[field.name] = (field_type, field.default)
            elif field.default_factory is not dataclasses.MISSING:
                field_definitions[field.name] = (field_type, field.default_factory)
            else:
                field_definitions[field.name] = (field_type, ...)

        model_dataclass_base: Any = create_model(output_type.__name__, **cast(Any, field_definitions))
        model = cast(type[BaseModel], model_dataclass_base)

        def _str_with_class_name(self: BaseModel) -> str:
            return repr(self)

        model.__str__ = _str_with_class_name  # type: ignore
        model._hypertic_str_overridden = True  # type: ignore
        return model

    is_typed_dict = (
        hasattr(output_type, "__annotations__")
        and not dataclasses.is_dataclass(output_type)
        and not issubclass(output_type, BaseModel)
        and (
            hasattr(output_type, "__total__")
            or hasattr(output_type, "__required_keys__")
            or hasattr(output_type, "__optional_keys__")
            or output_type.__module__ in ("typing", "typing_extensions")
        )
    )

    if is_typed_dict:
        try:
            hints = get_type_hints(output_type, include_extras=True)
            if hints:
                field_definitions = {}

                required_keys = set()
                if hasattr(output_type, "__required_keys__"):
                    required_keys = output_type.__required_keys__
                elif hasattr(output_type, "__total__"):
                    if output_type.__total__:
                        required_keys = set(hints.keys())

                for field_name, field_type in hints.items():
                    is_required = field_name in required_keys if required_keys else True

                    if is_required:
                        field_definitions[field_name] = (field_type, ...)
                    else:
                        if get_origin(field_type) is not type(None):
                            field_type = Optional[field_type]
                        field_definitions[field_name] = (field_type, None)

                model_dataclass_typed: Any = create_model(output_type.__name__, **cast(Any, field_definitions))
                model = cast(type[BaseModel], model_dataclass_typed)

                def _str_with_class_name(self: BaseModel) -> str:
                    return repr(self)

                model.__str__ = _str_with_class_name  # type: ignore
                model._hypertic_str_overridden = True  # type: ignore
                return model
        except Exception:
            pass

    try:
        hints = get_type_hints(output_type)
        if hints:
            field_definitions = {name: (hint, ...) for name, hint in hints.items()}
            model_dataclass_fallback: Any = create_model(output_type.__name__, **cast(Any, field_definitions))
            model = cast(type[BaseModel], model_dataclass_fallback)

            def _str_with_class_name(self: BaseModel) -> str:
                return repr(self)

            model.__str__ = _str_with_class_name  # type: ignore
            model._hypertic_str_overridden = True  # type: ignore
            return model
    except Exception:
        pass

    raise ValueError(f"Cannot convert {output_type} to Pydantic model. Supported types: Pydantic BaseModel, dataclass, or TypedDict.")


@dataclass(init=False)
class Agent:
    model: Any
    instructions: str | None = None
    tools: list[Any] | None = None
    max_steps: int = 10
    output_type: type[Any] | None = None
    parallel_calls: bool = True
    retriever: Any | None = None
    memory: Any | None = None
    guardrails: list[Any] | None = None
    handler: Any = field(default=None, init=False, repr=False)
    mcp_tools: list[Any] = field(default_factory=list, init=False, repr=False)
    function_tools: list[Any] = field(default_factory=list, init=False, repr=False)
    _tool_manager: Any = field(default=None, init=False, repr=False)
    file_processor: FileProcessor = field(default_factory=FileProcessor, init=False, repr=False)
    _tool_outputs: dict[str, Any] = field(default_factory=dict, init=False, repr=False)
    _tool_calls: list[Any] = field(default_factory=list, init=False, repr=False)

    def __init__(
        self,
        *,
        model: Any,
        instructions: str | None = None,
        tools: list[Any] | None = None,
        max_steps: int = 10,
        output_type: type[Any] | None = None,
        parallel_calls: bool = True,
        retriever: Any | None = None,
        memory: Any | None = None,
        guardrails: list[Any] | None = None,
    ):
        self.model = model
        self.instructions = instructions
        self.tools = tools
        self.max_steps = max_steps
        self.output_type = output_type
        self.parallel_calls = parallel_calls
        self.retriever = retriever
        self.memory = memory
        self.guardrails = guardrails or []

        self.handler = None
        self.mcp_tools = []
        self.function_tools = []
        self._tool_manager = None
        self.file_processor = FileProcessor()
        self._tool_outputs = {}
        self._tool_calls = []

        self._initialize()

    def _initialize(self) -> None:
        if self.tools is None:
            self.tools = []

        if not isinstance(self.parallel_calls, bool):
            raise ConfigurationError("parallel_calls must be True or False")

        if hasattr(self.model, "get_handler"):
            self.handler = self.model.get_handler()
        else:
            self.handler = self.model

        self._separate_tool_types()

        if self.function_tools:
            self._tool_manager = _ToolManager(self.function_tools)
        else:
            self._tool_manager = None

        if self.output_type is not None:
            provider_name = self.handler.__class__.__name__
            try:
                _convert_to_pydantic(self.output_type)
            except Exception as e:
                raise SchemaConversionError(f"Structured output not supported for {provider_name}: {e}") from e

    def _separate_tool_types(self):
        self.mcp_tools = []
        self.function_tools = []

        if self.tools is None:
            return

        all_tools: list[Any] = []

        def process_tool(tool: Any) -> None:
            if isinstance(tool, list):
                for item in tool:
                    process_tool(item)
            elif isinstance(tool, BaseToolkit):
                toolkit_tools = tool.get_tools()
                for toolkit_tool in toolkit_tools:
                    all_tools.append(toolkit_tool)
            else:
                all_tools.append(tool)

        if isinstance(self.tools, list):
            for tool in self.tools:
                process_tool(tool)
        else:
            process_tool(self.tools)

        for tool in all_tools:
            if hasattr(tool, "_tool_metadata"):
                self.function_tools.append(tool)
            elif isinstance(tool, ExecutableTool):
                self.mcp_tools.append(tool)
            elif isinstance(tool, type):
                raise ValueError(f"Tool must be an instance, not a class: {tool.__name__}.")
            elif hasattr(tool, "name") and hasattr(tool, "inputSchema"):
                self.mcp_tools.append(tool)
            elif isinstance(tool, (str, int, float, bool, list, dict)):
                continue
            elif hasattr(tool, "__dict__"):
                self.mcp_tools.append(tool)
            else:
                self.mcp_tools.append(tool)

    async def _get_all_tools_async(self) -> list[dict[str, Any]]:
        openai_tools = []

        if self.function_tools and self._tool_manager:
            openai_tools.extend(self._tool_manager.to_openai_format())

        for tool in self.mcp_tools:
            if isinstance(tool, dict) and "function" in tool:
                openai_tools.append(tool)
            elif hasattr(tool, "name"):
                openai_tool = {
                    "type": "function",
                    "function": {
                        "name": tool.name,
                        "description": getattr(tool, "description", ""),
                        "parameters": getattr(tool, "inputSchema", {}),
                    },
                }
                openai_tools.append(openai_tool)

        return openai_tools

    def _load_memory_context(self, user_id: str | None, session_id: str | None) -> list[dict[str, Any]]:
        messages: list[dict[str, Any]] = []

        if not self.memory:
            return messages

        if session_id and not user_id:
            raise ValueError(
                "session_id requires user_id to be provided. "
                "For session-based memory, both session_id and user_id are required. "
                "For long-term memory, provide only user_id."
            )

        try:
            if session_id:
                messages = self.memory.get_messages(session_id=session_id, user_id=user_id)
            elif user_id:
                messages = self.memory.get_messages(session_id=None, user_id=user_id)
        except Exception as e:
            logger.warning(f"Error loading messages: {e}")

        cleaned_messages = []
        for msg in messages:
            cleaned_msg = {"role": msg.get("role"), "content": msg.get("content")}
            cleaned_messages.append(cleaned_msg)

        return cleaned_messages

    async def _aload_memory_context(self, user_id: str | None, session_id: str | None) -> list[dict[str, Any]]:
        messages: list[dict[str, Any]] = []

        if not self.memory:
            return messages

        if session_id and not user_id:
            raise ValueError(
                "session_id requires user_id to be provided. "
                "For session-based memory, both session_id and user_id are required. "
                "For long-term memory, provide only user_id."
            )

        try:
            if session_id:
                messages = await self.memory.aget_messages(session_id=session_id, user_id=user_id)
            elif user_id:
                messages = await self.memory.aget_messages(session_id=None, user_id=user_id)
        except Exception as e:
            logger.warning(f"Error loading messages: {e}", exc_info=True)

        cleaned_messages = []
        for msg in messages:
            cleaned_msg = {"role": msg.get("role"), "content": msg.get("content")}
            cleaned_messages.append(cleaned_msg)

        return cleaned_messages

    def _save_to_memory(
        self,
        session_id: str,
        user_message: str,
        assistant_message: str,
        user_id: str | None = None,
        tool_calls: list[dict[str, Any]] | None = None,
        tool_outputs: dict[str, Any] | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> None:
        if not self.memory:
            return

        try:
            self.memory.save_message(session_id=session_id, role="user", content=user_message, user_id=user_id)

            self.memory.save_message(
                session_id=session_id,
                role="assistant",
                content=assistant_message,
                user_id=user_id,
                tool_calls=tool_calls,
                tool_outputs=tool_outputs,
                metadata=metadata,
            )
        except Exception as e:
            logger.warning(f"Error saving to memory: {e}")

    async def _asave_to_memory(
        self,
        session_id: str,
        user_message: str,
        assistant_message: str,
        user_id: str | None = None,
        tool_calls: list[dict[str, Any]] | None = None,
        tool_outputs: dict[str, Any] | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> None:
        if not self.memory:
            return

        try:
            await self.memory.asave_message(session_id=session_id, role="user", content=user_message, user_id=user_id)

            await self.memory.asave_message(
                session_id=session_id,
                role="assistant",
                content=assistant_message,
                user_id=user_id,
                tool_calls=tool_calls,
                tool_outputs=tool_outputs,
                metadata=metadata,
            )
        except Exception as e:
            logger.warning(f"Error saving to memory: {e}", exc_info=True)

    async def _retrieve_knowledge_async(self, query: str) -> str:
        if not self.retriever:
            return ""

        try:
            if hasattr(self.retriever, "async_search"):
                docs = await self.retriever.async_search(query)
                results = [
                    type(
                        "RetrievalResult",
                        (),
                        {
                            "content": doc.page_content,
                            "metadata": doc.metadata,
                            "score": getattr(doc, "score", 1.0),
                            "source": doc.metadata.get("source", "unknown"),
                        },
                    )()
                    for doc in docs
                ]
                return "\n\n".join([result.content for result in results])
            else:
                error_msg = f"Unknown knowledge source type: {type(self.retriever)}"
                logger.warning(error_msg)
                raise RetrieverError(error_msg)
        except RetrieverError:
            raise
        except Exception as e:
            logger.error(f"Error retrieving knowledge: {e}", exc_info=True)
            raise RetrieverError(f"Failed to retrieve knowledge: {e}") from e

    async def _avalidate_input(self, query: str, user_id: str | None = None, session_id: str | None = None) -> str:
        if not self.guardrails:
            return query

        modified_query = query
        for guardrail in self.guardrails:
            result = guardrail.validate_input(query=modified_query, user_id=user_id, session_id=session_id)

            if not result.allowed:
                raise GuardrailViolationError(
                    reason=result.reason or "Input blocked by guardrail",
                    violation_type=result.violation_type,
                    details=result.details,
                )

            if result.modified_input:
                modified_query = result.modified_input

        return modified_query

    def _validate_input(self, query: str, user_id: str | None = None, session_id: str | None = None) -> str:
        if not self.guardrails:
            return query

        modified_query = query
        for guardrail in self.guardrails:
            result = guardrail.validate_input(query=modified_query, user_id=user_id, session_id=session_id)

            if not result.allowed:
                raise GuardrailViolationError(
                    reason=result.reason or "Input blocked by guardrail",
                    violation_type=result.violation_type,
                    details=result.details,
                )

            if result.modified_input:
                modified_query = result.modified_input

        return modified_query

    async def arun(
        self,
        query: str,
        files: list[str] | None = None,
        user_id: str | None = None,
        session_id: str | None = None,
    ) -> LLMResponse:
        """Asynchronous run method - industry standard approach.

        Args:
            query: User query
            files: Optional list of file paths
            user_id: User identifier (required if session_id is provided, optional for long-term memory)
            session_id: Session identifier (requires user_id to be provided). Auto-generated if not provided.

        Returns:
            LLMResponse with the model's response

        Note:
            Memory behavior:
            - If session_id is provided → user_id is REQUIRED (short-term memory, session-specific)
            - If only user_id is provided (no session_id) → long-term memory (all user messages)
            - Only session_id (without user_id) → NOT ALLOWED (raises ValueError)
        """
        try:
            query = await self._avalidate_input(query, user_id=user_id, session_id=session_id)

            self._tool_calls = []
            self._tool_outputs = {}

            memory_messages = await self._aload_memory_context(
                user_id=user_id,
                session_id=session_id,
            )

            if session_id is None:
                session_id = f"session_{uuid.uuid4().hex[:8]}"

            if self.retriever:
                context = await self._retrieve_knowledge_async(query)
                if context:
                    query = f"{query}\n\n{context}"

            non_stream_response: LLMResponse = await self._arun_non_streaming(
                query=query, files=files, memory_messages=memory_messages, session_id=session_id
            )

            if non_stream_response.content:
                tool_calls: list[dict[str, Any]] | None = self._tool_calls if self._tool_calls else None
                tool_outputs: dict[str, Any] | None = self._tool_outputs if self._tool_outputs else None
                llm_metadata: dict[str, Any] | None = non_stream_response.metadata if non_stream_response.metadata else None
                await self._asave_to_memory(
                    session_id,
                    query,
                    non_stream_response.content,
                    user_id=user_id,
                    tool_calls=tool_calls,
                    tool_outputs=tool_outputs,
                    metadata=llm_metadata,
                )

            if self.output_type is not None and non_stream_response.content:
                try:
                    import json

                    if isinstance(non_stream_response.content, str):
                        parsed_dict = json.loads(non_stream_response.content)
                        pydantic_model = _convert_to_pydantic(self.output_type)
                        structured_output = pydantic_model(**parsed_dict)
                        if not hasattr(structured_output.__class__, "_hypertic_str_overridden"):

                            def _str_with_class_name(self: BaseModel) -> str:
                                return repr(self)

                            structured_output.__class__.__str__ = _str_with_class_name  # type: ignore
                            structured_output.__class__._hypertic_str_overridden = True  # type: ignore
                        object.__setattr__(non_stream_response, "structured_output", structured_output)
                        object.__setattr__(non_stream_response, "content", "")
                except (json.JSONDecodeError, ValueError, TypeError):
                    pass

            return non_stream_response
        finally:
            await self._auto_disconnect_clients()

    async def astream(
        self,
        query: str,
        files: list[str] | None = None,
        user_id: str | None = None,
        session_id: str | None = None,
    ) -> AsyncIterator[StreamEvent]:
        """Asynchronous streaming run method - industry standard approach.

        Args:
            query: User query
            files: Optional list of file paths
            user_id: User identifier (required if session_id is provided, optional for long-term memory)
            session_id: Session identifier (requires user_id to be provided). Auto-generated if not provided.

        Note:
            Memory behavior:
            - If session_id is provided → user_id is REQUIRED (short-term memory, session-specific)
            - If only user_id is provided (no session_id) → long-term memory (all user messages)
            - Only session_id (without user_id) → NOT ALLOWED (raises ValueError)
        """
        try:
            memory_messages = await self._aload_memory_context(
                user_id=user_id,
                session_id=session_id,
            )

            if session_id is None:
                session_id = f"session_{uuid.uuid4().hex[:8]}"

            if self.retriever:
                try:
                    context = await self._retrieve_knowledge_async(query)
                    if context:
                        query = f"{query}\n\n{context}"
                except RetrieverError as e:
                    logger.warning(f"Continuing without knowledge retrieval: {e}")

            response_content = ""
            collected_tool_calls: list[dict[str, Any]] = []
            collected_tool_outputs: dict[str, Any] = {}
            collected_metadata: dict[str, Any] | None = None

            async for chunk in self._arun_streaming(query=query, files=files, memory_messages=memory_messages):
                if chunk.type == "content":
                    response_content += chunk.content
                    if self.output_type is None:
                        yield chunk
                elif chunk.type == "tool_calls":
                    if chunk.tool_calls:
                        collected_tool_calls.extend(chunk.tool_calls)
                    yield chunk
                elif chunk.type == "tool_outputs":
                    if chunk.tool_outputs:
                        collected_tool_outputs.update(chunk.tool_outputs)
                    yield chunk
                elif chunk.type == "metadata":
                    collected_metadata = chunk.metadata
                else:
                    yield chunk

            if self.output_type is not None and response_content:
                try:
                    import json

                    from hypertic.models.events import StructuredOutputEvent

                    if isinstance(response_content, str):
                        parsed_dict = json.loads(response_content)
                        pydantic_model = _convert_to_pydantic(self.output_type)
                        structured_output = pydantic_model(**parsed_dict)
                        if not hasattr(structured_output.__class__, "_hypertic_str_overridden"):

                            def _str_with_class_name(self: BaseModel) -> str:
                                return repr(self)

                            structured_output.__class__.__str__ = _str_with_class_name  # type: ignore
                            structured_output.__class__._hypertic_str_overridden = True  # type: ignore
                        yield StructuredOutputEvent(structured_output=structured_output)
                except (json.JSONDecodeError, ValueError, TypeError):
                    pass

            if collected_metadata is not None:
                from hypertic.models.events import MetadataEvent

                yield MetadataEvent(metadata=collected_metadata)

            if response_content and self.memory:
                await self._asave_to_memory(
                    session_id,
                    query,
                    response_content,
                    user_id=user_id,
                    tool_calls=collected_tool_calls if collected_tool_calls else None,
                    tool_outputs=collected_tool_outputs if collected_tool_outputs else None,
                    metadata=collected_metadata,
                )
        finally:
            await self._auto_disconnect_clients()

    async def _arun_non_streaming(
        self,
        query: str,
        files: list[str] | None = None,
        memory_messages: list[dict[str, Any]] | None = None,
        session_id: str | None = None,
    ) -> LLMResponse:
        messages: list[dict[str, Any]] = []

        if self.instructions is not None and self.instructions.strip():
            messages.append({"role": "system", "content": self.instructions})

        user_message: dict[str, Any] = {"role": "user", "content": query}
        if files:
            user_message["files"] = files

        if files:
            provider_name = self.handler.__class__.__name__
            user_message = self.file_processor.process_message(user_message, provider_name)

        messages.append(user_message)

        openai_tools = await self._get_all_tools_async()

        result: LLMResponse = await self._ahandle_non_streaming_with_tools(messages, openai_tools, memory_messages=memory_messages)
        return result

    async def _arun_streaming(
        self,
        query: str,
        files: list[str] | None = None,
        memory_messages: list[dict[str, Any]] | None = None,
    ):
        messages: list[dict[str, Any]] = []

        if self.instructions is not None and self.instructions.strip():
            messages.append({"role": "system", "content": self.instructions})

        user_message: dict[str, Any] = {"role": "user", "content": query}
        if files:
            user_message["files"] = files

        if files:
            provider_name = self.handler.__class__.__name__
            user_message = self.file_processor.process_message(user_message, provider_name)

        messages.append(user_message)

        openai_tools = await self._get_all_tools_async()

        async for chunk in self._ahandle_streaming_with_tools(messages, openai_tools, memory_messages=memory_messages):
            yield chunk

    async def _aexecute_tool(
        self,
        tool_name: str,
        arguments: dict[str, Any],
        user_id: str | None = None,
        session_id: str | None = None,
    ) -> str:
        if self.function_tools and self._tool_manager:
            try:
                loop = asyncio.get_event_loop()
                result: str = await loop.run_in_executor(None, self._tool_manager.execute_tool, tool_name, arguments)
                self._tool_outputs[tool_name] = result
                return result
            except ToolNotFoundError:
                pass
            except Exception as e:
                logger.error(f"Error executing function tool '{tool_name}': {e}", exc_info=True)
                raise ToolExecutionError(f"Failed to execute function tool '{tool_name}': {e}") from e

        for tool in self.mcp_tools:
            if hasattr(tool, "name") and tool.name == tool_name:
                if hasattr(tool, "call_tool"):
                    try:
                        if asyncio.iscoroutinefunction(tool.call_tool):
                            result = await tool.call_tool(arguments)
                        else:
                            result = tool.call_tool(arguments)

                        self._tool_outputs[tool_name] = str(result)
                        return str(result)
                    except Exception as e:
                        logger.error(f"Error executing tool '{tool_name}': {e}", exc_info=True)
                        raise ToolExecutionError(f"Failed to execute tool '{tool_name}': {e}") from e
                else:
                    result = f"Tool '{tool_name}' executed with arguments: {arguments}"
                    self._tool_outputs[tool_name] = result
                    return result

        raise ToolNotFoundError(f"Tool '{tool_name}' not found")

    async def _aexecute_tools_parallel(
        self,
        tool_calls: list[dict[str, Any]],
        user_id: str | None = None,
        session_id: str | None = None,
    ) -> dict[str, str]:
        async def execute_single_tool(tool_call):
            function_name = tool_call["function"]["name"]
            try:
                function_args = json.loads(tool_call["function"]["arguments"])
                if function_args is None:
                    function_args = {}
            except (json.JSONDecodeError, TypeError):
                function_args = {}

            result = await self._aexecute_tool(function_name, function_args, user_id=user_id, session_id=session_id)
            return function_name, result

        tasks = [execute_single_tool(tool_call) for tool_call in tool_calls]
        results = await asyncio.gather(*tasks, return_exceptions=True)

        tool_outputs: dict[str, Any] = {}
        for result in results:
            if isinstance(result, Exception):
                if isinstance(result, ToolExecutionError | ToolNotFoundError):
                    raise result
                logger.error(f"Unexpected error during tool execution: {result}", exc_info=True)
                raise ToolExecutionError(f"Unexpected error during tool execution: {result}") from result
            else:
                if isinstance(result, tuple) and len(result) == 2:
                    function_name, tool_result = result
                    tool_outputs[function_name] = tool_result
                else:
                    raise ToolExecutionError(f"Unexpected tool result: {result!r}")

        return tool_outputs

    async def _auto_disconnect_clients(self):
        for tool in self.mcp_tools:
            if hasattr(tool, "mcp_servers") and tool.mcp_servers.initialized:
                await tool.mcp_servers.disconnect()

    async def _ahandle_non_streaming_with_tools(
        self,
        messages: list[dict[str, Any]],
        tools: list[dict[str, Any]] | None,
        memory_messages: list[dict[str, Any]] | None = None,
    ) -> Any:
        self.handler._reset_metrics()

        if memory_messages:
            messages = memory_messages + messages

        response_format = None
        if self.output_type is not None:
            try:
                response_format = _convert_to_pydantic(self.output_type)
            except Exception as e:
                logger.warning(
                    f"Could not convert structured output type: {e}",
                    exc_info=True,
                )
                response_format = None

        for _step in range(self.max_steps):
            response = await self.handler.ahandle_non_streaming(self.handler, messages, tools, self, response_format)

            if response is None:
                continue

            return cast(LLMResponse, response)

        raise MaxStepsError(f"Maximum steps ({self.max_steps}) reached without completion")

    async def _ahandle_streaming_with_tools(
        self,
        messages: list[dict[str, Any]],
        tools: list[dict[str, Any]] | None,
        memory_messages: list[dict[str, Any]] | None = None,
    ):
        self.handler._reset_metrics()

        if memory_messages:
            messages = memory_messages + messages

        response_format = None
        if self.output_type is not None:
            try:
                response_format = _convert_to_pydantic(self.output_type)
            except Exception as e:
                logger.warning(
                    f"Could not convert structured output type: {e}",
                    exc_info=True,
                )
                response_format = None

        finish_reason = "stop"

        for _step in range(self.max_steps):
            has_more_tools = False

            async for event in self.handler.ahandle_streaming(self.handler, messages, tools, self, response_format):
                if isinstance(event, bool):
                    has_more_tools = event
                else:
                    if event.type == "tool_calls":
                        yield event
                        has_more_tools = True
                        continue
                    elif event.type == "tool_outputs":
                        yield event
                        has_more_tools = True
                        break
                    else:
                        yield event

            if not has_more_tools:
                break

        cumulative_metrics = self.handler._get_cumulative_metrics()

        params = {}
        if self.handler.temperature is not None:
            params["temperature"] = self.handler.temperature
        if self.handler.top_p is not None:
            params["top_p"] = self.handler.top_p
        if self.handler.presence_penalty is not None:
            params["presence_penalty"] = self.handler.presence_penalty
        if self.handler.frequency_penalty is not None:
            params["frequency_penalty"] = self.handler.frequency_penalty
        if self.handler.max_tokens is not None:
            params["max_tokens"] = self.handler.max_tokens

        final_metadata = {
            "model": self.handler.model,
            "params": params,
            "finish_reason": finish_reason,
            "input_tokens": cumulative_metrics.input_tokens,
            "output_tokens": cumulative_metrics.output_tokens,
        }

        yield MetadataEvent(metadata=final_metadata)

    def _get_all_tools_sync(self) -> list[dict[str, Any]]:
        openai_tools = []

        if self.function_tools and self._tool_manager:
            openai_tools.extend(self._tool_manager.to_openai_format())

        return openai_tools

    def _retrieve_knowledge_sync(self, query: str) -> str:
        if not self.retriever:
            return ""

        try:
            if hasattr(self.retriever, "search"):
                docs = self.retriever.search(query)
                results = [
                    type(
                        "RetrievalResult",
                        (),
                        {
                            "content": doc.page_content,
                            "metadata": doc.metadata,
                            "score": getattr(doc, "score", 1.0),
                            "source": doc.metadata.get("source", "unknown"),
                        },
                    )()
                    for doc in docs
                ]
                return "\n\n".join([result.content for result in results])
            else:
                error_msg = f"Unknown knowledge source type: {type(self.retriever)}"
                logger.warning(error_msg)
                raise RetrieverError(error_msg)
        except RetrieverError:
            raise
        except Exception as e:
            logger.error(f"Error retrieving knowledge: {e}", exc_info=True)
            raise RetrieverError(f"Failed to retrieve knowledge: {e}") from e

    def run(
        self,
        query: str,
        files: list[str] | None = None,
        user_id: str | None = None,
        session_id: str | None = None,
    ) -> LLMResponse:
        """Synchronous run method - industry standard approach.

        Args:
            query: User query
            files: Optional list of file paths
            user_id: User identifier (required if session_id is provided, optional for long-term memory)
            session_id: Session identifier (requires user_id to be provided). Auto-generated if not provided.

        Returns:
            LLMResponse with content and metadata

        Note:
            Memory behavior:
            - If session_id is provided → user_id is REQUIRED (short-term memory, session-specific)
            - If only user_id is provided (no session_id) → long-term memory (all user messages)
            - Only session_id (without user_id) → NOT ALLOWED (raises ValueError)
        """
        query = self._validate_input(query, user_id=user_id, session_id=session_id)

        self._tool_calls = []
        self._tool_outputs = {}

        memory_messages = self._load_memory_context(
            user_id=user_id,
            session_id=session_id,
        )

        if session_id is None:
            session_id = f"session_{uuid.uuid4().hex[:8]}"

        if self.retriever:
            try:
                context = self._retrieve_knowledge_sync(query)
                if context:
                    query = f"{query}\n\n{context}"
            except RetrieverError as e:
                logger.warning(f"Continuing without knowledge retrieval: {e}")

        response: LLMResponse = self._run_non_streaming(query=query, files=files, memory_messages=memory_messages, session_id=session_id)

        if response.content:
            tool_calls = self._tool_calls if self._tool_calls else None
            tool_outputs = self._tool_outputs if self._tool_outputs else None
            metadata = response.metadata if response.metadata else None
            self._save_to_memory(
                session_id,
                query,
                response.content,
                user_id=user_id,
                tool_calls=tool_calls,
                tool_outputs=tool_outputs,
                metadata=metadata,
            )

        if self.output_type is not None and response.content:
            try:
                import json

                if isinstance(response.content, str):
                    parsed_dict = json.loads(response.content)
                    pydantic_model = _convert_to_pydantic(self.output_type)
                    structured_output = pydantic_model(**parsed_dict)
                    if not hasattr(structured_output.__class__, "_hypertic_str_overridden"):

                        def _str_with_class_name(self: BaseModel) -> str:
                            return repr(self)

                        structured_output.__class__.__str__ = _str_with_class_name  # type: ignore
                        structured_output.__class__._hypertic_str_overridden = True  # type: ignore
                    object.__setattr__(response, "structured_output", structured_output)
                    object.__setattr__(response, "content", "")
            except (json.JSONDecodeError, ValueError, TypeError):
                pass

        return response

    def stream(
        self,
        query: str,
        files: list[str] | None = None,
        user_id: str | None = None,
        session_id: str | None = None,
    ) -> Iterator[StreamEvent]:
        """Synchronous streaming run method - industry standard approach.

        Args:
            query: User query
            files: Optional list of file paths
            user_id: User identifier (required if session_id is provided, optional for long-term memory)
            session_id: Session identifier (requires user_id to be provided). Auto-generated if not provided.

        Note:
            Memory behavior:
            - If session_id is provided → user_id is REQUIRED (short-term memory, session-specific)
            - If only user_id is provided (no session_id) → long-term memory (all user messages)
            - Only session_id (without user_id) → NOT ALLOWED (raises ValueError)
        """
        memory_messages = self._load_memory_context(
            user_id=user_id,
            session_id=session_id,
        )

        if session_id is None:
            session_id = f"session_{uuid.uuid4().hex[:8]}"

        if self.retriever:
            try:
                context = self._retrieve_knowledge_sync(query)
                if context:
                    query = f"{query}\n\n{context}"
            except RetrieverError as e:
                logger.warning(f"Continuing without knowledge retrieval: {e}")

        response_content = ""
        collected_tool_calls: list[dict[str, Any]] = []
        collected_tool_outputs: dict[str, Any] = {}
        collected_metadata: dict[str, Any] | None = None

        for chunk in self._run_streaming(query=query, files=files, memory_messages=memory_messages):
            if chunk.type == "content":
                response_content += chunk.content
                if self.output_type is None:
                    yield chunk
            elif chunk.type == "tool_calls":
                if chunk.tool_calls:
                    collected_tool_calls.extend(chunk.tool_calls)
                yield chunk
            elif chunk.type == "tool_outputs":
                if chunk.tool_outputs:
                    collected_tool_outputs.update(chunk.tool_outputs)
                yield chunk
            elif chunk.type == "metadata":
                collected_metadata = chunk.metadata
            else:
                yield chunk

        if self.output_type is not None and response_content:
            try:
                import json

                from hypertic.models.events import StructuredOutputEvent

                if isinstance(response_content, str):
                    parsed_dict = json.loads(response_content)
                    pydantic_model = _convert_to_pydantic(self.output_type)
                    structured_output = pydantic_model(**parsed_dict)
                    if not hasattr(structured_output.__class__, "_hypertic_str_overridden"):

                        def _str_with_class_name(self: BaseModel) -> str:
                            return repr(self)

                        structured_output.__class__.__str__ = _str_with_class_name  # type: ignore
                        structured_output.__class__._hypertic_str_overridden = True  # type: ignore
                    yield StructuredOutputEvent(structured_output=structured_output)
            except (json.JSONDecodeError, ValueError, TypeError):
                pass

        if collected_metadata is not None:
            from hypertic.models.events import MetadataEvent

            yield MetadataEvent(metadata=collected_metadata)

        if response_content and self.memory:
            self._save_to_memory(
                session_id,
                query,
                response_content,
                user_id=user_id,
                tool_calls=collected_tool_calls if collected_tool_calls else None,
                tool_outputs=collected_tool_outputs if collected_tool_outputs else None,
                metadata=collected_metadata,
            )

    def _run_non_streaming(
        self,
        query: str,
        files: list[str] | None = None,
        memory_messages: list[dict[str, Any]] | None = None,
        session_id: str | None = None,
    ) -> LLMResponse:
        messages: list[dict[str, Any]] = []
        if self.instructions is not None and self.instructions.strip():
            messages.append({"role": "system", "content": self.instructions})

        user_message: dict[str, Any] = {"role": "user", "content": query}
        if files:
            user_message["files"] = files

        if files:
            provider_name = self.handler.__class__.__name__
            user_message = self.file_processor.process_message(user_message, provider_name)

        messages.append(user_message)

        openai_tools = self._get_all_tools_sync()

        return cast(
            LLMResponse,
            self._handle_non_streaming_with_tools(messages, openai_tools, memory_messages=memory_messages),
        )

    def _run_streaming(
        self,
        query: str,
        files: list[str] | None = None,
        memory_messages: list[dict[str, Any]] | None = None,
    ):
        messages: list[dict[str, Any]] = []
        if self.instructions is not None and self.instructions.strip():
            messages.append({"role": "system", "content": self.instructions})

        user_message: dict[str, Any] = {"role": "user", "content": query}
        if files:
            user_message["files"] = files

        if files:
            provider_name = self.handler.__class__.__name__
            user_message = self.file_processor.process_message(user_message, provider_name)

        messages.append(user_message)

        openai_tools = self._get_all_tools_sync()

        yield from self._handle_streaming_with_tools(messages, openai_tools, memory_messages=memory_messages)

    def _execute_tool(
        self,
        tool_name: str,
        arguments: dict[str, Any],
        user_id: str | None = None,
        session_id: str | None = None,
    ) -> str:
        if self.function_tools and self._tool_manager:
            try:
                result: str = str(self._tool_manager.execute_tool(tool_name, arguments))
                self._tool_outputs[tool_name] = result
                return result
            except ToolNotFoundError:
                pass
            except Exception as e:
                logger.error(f"Error executing function tool '{tool_name}': {e}", exc_info=True)
                raise ToolExecutionError(f"Failed to execute function tool '{tool_name}': {e}") from e

        for tool in self.mcp_tools:
            if hasattr(tool, "name") and tool.name == tool_name:
                if hasattr(tool, "mcp_servers"):
                    raise ToolExecutionError(
                        f"Tool '{tool_name}' is an MCP tool (async-only). Use agent.arun() or agent.astream() instead of run() or stream()."
                    )

                if hasattr(tool, "call_tool"):
                    try:
                        result = tool.call_tool(arguments)
                        self._tool_outputs[tool_name] = str(result)
                        return str(result)
                    except Exception as e:
                        logger.error(f"Error executing tool '{tool_name}': {e}", exc_info=True)
                        raise ToolExecutionError(f"Failed to execute tool '{tool_name}': {e}") from e
                else:
                    result = f"Tool '{tool_name}' executed with arguments: {arguments}"
                    self._tool_outputs[tool_name] = result
                    return result

        raise ToolNotFoundError(f"Tool '{tool_name}' not found")

    def _execute_tools_parallel(
        self,
        tool_calls: list[dict[str, Any]],
        user_id: str | None = None,
        session_id: str | None = None,
    ) -> dict[str, str]:
        def execute_single_tool(tool_call):
            function_name = tool_call["function"]["name"]
            try:
                function_args = json.loads(tool_call["function"]["arguments"])
                if function_args is None:
                    function_args = {}
            except (json.JSONDecodeError, TypeError):
                function_args = {}

            result = self._execute_tool(function_name, function_args, user_id=user_id, session_id=session_id)
            return function_name, result

        results = [execute_single_tool(tool_call) for tool_call in tool_calls]

        tool_outputs = {}
        for result in results:
            if isinstance(result, Exception):
                if isinstance(result, ToolExecutionError | ToolNotFoundError):
                    raise result
                logger.error(f"Unexpected error during tool execution: {result}", exc_info=True)
                raise ToolExecutionError(f"Unexpected error during tool execution: {result}") from result
            else:
                function_name, tool_result = result
                tool_outputs[function_name] = tool_result

        return tool_outputs

    def _handle_non_streaming_with_tools(
        self,
        messages: list[dict[str, Any]],
        tools: list[dict[str, Any]] | None,
        memory_messages: list[dict[str, Any]] | None = None,
    ) -> Any:
        self.handler._reset_metrics()

        if memory_messages:
            messages = memory_messages + messages

        response_format = None
        if self.output_type is not None:
            try:
                response_format = _convert_to_pydantic(self.output_type)
            except Exception as e:
                logger.warning(
                    f"Could not convert structured output type: {e}",
                    exc_info=True,
                )
                response_format = None

        for _step in range(self.max_steps):
            response: LLMResponse | None = self.handler.handle_non_streaming(self.handler, messages, tools, self, response_format)

            if response is None:
                continue

            return response

        raise MaxStepsError(f"Maximum steps ({self.max_steps}) reached without completion")

    def _handle_streaming_with_tools(
        self,
        messages: list[dict[str, Any]],
        tools: list[dict[str, Any]] | None,
        memory_messages: list[dict[str, Any]] | None = None,
    ):
        self.handler._reset_metrics()

        if memory_messages:
            messages = memory_messages + messages

        response_format = None
        if self.output_type is not None:
            try:
                response_format = _convert_to_pydantic(self.output_type)
            except Exception as e:
                logger.warning(
                    f"Could not convert structured output type: {e}",
                    exc_info=True,
                )
                response_format = None

        finish_reason = "stop"

        for _step in range(self.max_steps):
            has_more_tools = False

            for event in self.handler.handle_streaming(self.handler, messages, tools, self, response_format):
                if isinstance(event, bool):
                    has_more_tools = event
                else:
                    if event.type == "tool_calls":
                        yield event
                        has_more_tools = True
                        continue
                    elif event.type == "tool_outputs":
                        yield event
                        has_more_tools = True
                        break
                    else:
                        yield event

            if not has_more_tools:
                break

        cumulative_metrics = self.handler._get_cumulative_metrics()

        params = {}
        if self.handler.temperature is not None:
            params["temperature"] = self.handler.temperature
        if self.handler.top_p is not None:
            params["top_p"] = self.handler.top_p
        if self.handler.presence_penalty is not None:
            params["presence_penalty"] = self.handler.presence_penalty
        if self.handler.frequency_penalty is not None:
            params["frequency_penalty"] = self.handler.frequency_penalty
        if self.handler.max_tokens is not None:
            params["max_tokens"] = self.handler.max_tokens

        final_metadata = {
            "model": self.handler.model,
            "params": params,
            "finish_reason": finish_reason,
            "input_tokens": cumulative_metrics.input_tokens,
            "output_tokens": cumulative_metrics.output_tokens,
        }

        yield MetadataEvent(metadata=final_metadata)



import json
from typing import Optional

from hypertic.tools.base import BaseToolkit, tool


class DuckDuckGoTools(BaseToolkit):
    modifier: Optional[str] = None
    max_results: int = 5
    proxy: Optional[str] = None
    timeout: Optional[int] = 10
    backend: str = "auto"
    verify_ssl: bool = True
    region: Optional[str] = "us-en"
    safesearch: str = "moderate"
    time: Optional[str] = "y"

    @tool
    def search(self, query: str) -> str:
        """Search DuckDuckGo for a query.

        Uses the max_results configured during initialization (default: 5).

        Args:
            query: The query to search for.

        Returns:
            JSON string containing search results.
        """
        try:
            from ddgs import DDGS
        except ImportError as err:
            raise ImportError("ddgs library not installed. Install with: pip install ddgs") from err

        search_query = f"{self.modifier} {query}" if self.modifier else query

        try:
            with DDGS(proxy=self.proxy, timeout=self.timeout, verify=self.verify_ssl) as ddgs:
                results = list(
                    ddgs.text(
                        query=search_query,
                        region=self.region,
                        safesearch=self.safesearch,
                        timelimit=self.time,
                        max_results=self.max_results,
                        backend=self.backend,
                    )
                )

            return json.dumps(results, indent=2)
        except Exception as e:
            raise ValueError(f"Failed to search: {e!s}") from e

    @tool
    def news(self, query: str) -> str:
        """Get the latest news from DuckDuckGo.

        Uses the max_results configured during initialization (default: 5).

        Args:
            query: The query to search for.

        Returns:
            JSON string containing news results.
        """
        try:
            from ddgs import DDGS
        except ImportError as err:
            raise ImportError("ddgs library not installed. Install with: pip install ddgs") from err

        try:
            with DDGS(proxy=self.proxy, timeout=self.timeout, verify=self.verify_ssl) as ddgs:
                results = list(
                    ddgs.news(query=query, region=self.region, safesearch=self.safesearch, timelimit=self.time, max_results=self.max_results)
                )

            return json.dumps(results, indent=2)
        except Exception as e:
            raise ValueError(f"Failed to search news: {e!s}") from e

    @tool
    def images(self, query: str) -> str:
        """Search for images using DuckDuckGo.

        Uses the max_results configured during initialization (default: 5).

        Args:
            query: The query to search for.

        Returns:
            JSON string containing image search results.
        """
        try:
            from ddgs import DDGS
        except ImportError as err:
            raise ImportError("ddgs library not installed. Install with: pip install ddgs") from err

        try:
            with DDGS(proxy=self.proxy, timeout=self.timeout, verify=self.verify_ssl) as ddgs:
                results = list(ddgs.images(query=query, region=self.region, safesearch=self.safesearch, max_results=self.max_results))

            return json.dumps(results, indent=2)
        except Exception as e:
            raise ValueError(f"Failed to search images: {e!s}") from e



import json
from contextlib import AsyncExitStack
from dataclasses import dataclass, field
from typing import Any

from mcp import ClientSession, StdioServerParameters
from mcp.client.sse import sse_client
from mcp.client.stdio import stdio_client
from pydantic import AnyUrl

from hypertic.utils.log import get_logger

logger = get_logger(__name__)


def _get_streamable_http_client(url: str, headers: dict[str, str] | None, timeout: float, sse_read_timeout: float):  # noqa: ANN201
    """Get streamable HTTP client - wrapper for deprecated function.

    Note: streamablehttp_client is deprecated but has no replacement in current mcp version.
    This wrapper isolates the deprecation to a single location.
    """
    from mcp.client.streamable_http import streamablehttp_client  # noqa: PLC0415

    return streamablehttp_client(url, headers, timeout, sse_read_timeout)


@dataclass
class ExecutableTool:
    tool: Any
    mcp_servers: "MCPServers"
    name: str = field(init=False)
    description: str = field(init=False)
    input_schema: dict[str, Any] = field(init=False, default_factory=dict)

    def __post_init__(self):
        self.name = self.tool.name
        self.description = getattr(self.tool, "description", "")
        self.input_schema = getattr(self.tool, "inputSchema", {})

    def __repr__(self) -> str:
        tool_dict = {
            "name": self.name,
            "description": self.description,
            "parameters": self.input_schema,
        }
        return str(tool_dict)

    async def call_tool(self, arguments: dict[str, Any]) -> Any:
        return await self.mcp_servers.call_tool(self.name, arguments)


@dataclass
class MCPServerConfig:
    name: str
    transport: str
    config: dict[str, Any]


@dataclass
class MCPServers:
    config: dict[str, Any] | list[dict[str, Any]]

    servers: dict[str, MCPServerConfig] = field(default_factory=dict, init=False)
    sessions: dict[str, ClientSession] = field(default_factory=dict, init=False)
    exit_stack: AsyncExitStack = field(default_factory=AsyncExitStack, init=False)
    tools: dict[str, Any] = field(default_factory=dict, init=False)
    resources: dict[str, Any] = field(default_factory=dict, init=False)
    prompts: dict[str, Any] = field(default_factory=dict, init=False)
    initialized: bool = field(default=False, init=False)
    _auto_initialized: bool = field(default=False, init=False)

    def __post_init__(self):
        configs = [self.config] if isinstance(self.config, dict) else self.config

        for server_config in configs:
            self._add_server_config(server_config)

    def _add_server_config(self, config: dict[str, Any]):
        for server_name, server_config in config.items():
            transport = server_config.get("transport", "stdio")

            if transport == "stdio" and "command" in server_config:
                pass
            elif "url" in server_config:
                if transport == "streamable_http":
                    pass
                elif server_config.get("streamable", False):
                    transport = "streamable_http"
                else:
                    transport = "sse"

            self.servers[server_name] = MCPServerConfig(name=server_name, transport=transport, config=server_config)

    @classmethod
    def from_config_file(cls, *config_files: str, tools: list[str] | None = None) -> "MCPServers":
        configs = []
        for config_file in config_files:
            with open(config_file) as f:
                if config_file.lower().endswith((".yaml", ".yml")):
                    try:
                        import yaml

                        configs.append(yaml.safe_load(f))
                    except ImportError as err:
                        raise ImportError(
                            "YAML processing requires PyYAML. This should be installed by default. If missing, install with: pip install PyYAML"
                        ) from err
                else:
                    configs.append(json.load(f))
        return cls(configs)

    async def initialize(self) -> "MCPServers":
        if self.initialized:
            return self

        for server_name in self.servers:
            server_config = self.servers[server_name]
            session = await self._create_session(server_config)
            self.sessions[server_name] = session

            await self._discover_capabilities(server_name, session)

        self.initialized = True
        return self

    async def _create_session(self, server_config: MCPServerConfig) -> ClientSession:
        if server_config.transport == "stdio":
            return await self._create_stdio_session(server_config.config)
        elif server_config.transport == "sse":
            return await self._create_sse_session(server_config.config)
        elif server_config.transport == "streamable_http":
            return await self._create_streamable_http_session(server_config.config)
        else:
            raise ValueError(f"Unsupported transport type: {server_config.transport}")

    async def _create_stdio_session(self, config: dict[str, Any]) -> ClientSession:
        server_params = StdioServerParameters(
            command=config["command"],
            args=config.get("args", []),
            env=config.get("env", {}),
            cwd=config.get("cwd"),
        )

        stdio_transport = await self.exit_stack.enter_async_context(stdio_client(server_params))
        read_stream, write_stream = stdio_transport

        session = await self.exit_stack.enter_async_context(ClientSession(read_stream, write_stream))
        await session.initialize()

        return session

    async def _create_sse_session(self, config: dict[str, Any]) -> ClientSession:
        url = config["url"]
        headers = config.get("headers", {})
        timeout = config.get("timeout", 5)
        sse_read_timeout = config.get("sse_read_timeout", 60 * 5)

        sse_transport = await self.exit_stack.enter_async_context(sse_client(url, headers, timeout, sse_read_timeout))
        read_stream, write_stream = sse_transport

        session = await self.exit_stack.enter_async_context(ClientSession(read_stream, write_stream))
        await session.initialize()

        return session

    async def _create_streamable_http_session(self, config: dict[str, Any]) -> ClientSession:
        url = config["url"]
        headers = config.get("headers", {})
        timeout = config.get("timeout", 5)
        sse_read_timeout = config.get("sse_read_timeout", 60 * 5)

        # Use wrapper function to isolate deprecated import
        http_transport: tuple[Any, Any, Any] = await self.exit_stack.enter_async_context(
            _get_streamable_http_client(url, headers, timeout, sse_read_timeout)
        )
        read_stream, write_stream, _ = http_transport

        session = await self.exit_stack.enter_async_context(ClientSession(read_stream, write_stream))
        await session.initialize()

        return session

    async def _discover_capabilities(self, server_name: str, session: ClientSession):
        try:
            try:
                tools_response = await session.list_tools()

                tools_to_add = tools_response.tools

                for tool in tools_to_add:
                    self.tools[tool.name] = ExecutableTool(tool, self)

                logger.debug(f"Discovered {len(tools_to_add)} tools from {server_name} (filtered from {len(tools_response.tools)})")
            except Exception as e:
                logger.debug(f"Server {server_name} does not support list_tools: {e}")

            try:
                resources_response = await session.list_resources()
                for resource in resources_response.resources:
                    self.resources[str(resource.uri)] = resource
                logger.debug(f"Discovered {len(resources_response.resources)} resources from {server_name}")
            except Exception as e:
                logger.debug(f"Server {server_name} does not support list_resources: {e}")

            try:
                prompts_response = await session.list_prompts()
                for prompt in prompts_response.prompts:
                    self.prompts[prompt.name] = prompt
                logger.debug(f"Discovered {len(prompts_response.prompts)} prompts from {server_name}")
            except Exception as e:
                logger.debug(f"Server {server_name} does not support list_prompts: {e}")

        except Exception as e:
            logger.error(f"Failed to discover capabilities from {server_name}: {e}")

    async def call_tool(self, tool_name: str, arguments: dict[str, Any]) -> Any:
        if not self.initialized:
            await self.initialize()

        tool = self.tools.get(tool_name)
        if not tool:
            raise ValueError(f"Tool '{tool_name}' not found")

        try:
            server_name = None
            for name, session in self.sessions.items():
                tools_response = await session.list_tools()
                if any(t.name == tool_name for t in tools_response.tools):
                    server_name = name
                    break

            if not server_name:
                raise ValueError(f"Tool '{tool_name}' not found on any server")

            session = self.sessions[server_name]

            result = await session.call_tool(tool_name, arguments)

            if hasattr(result, "content") and result.content:
                content_parts = []
                for content in result.content:
                    if hasattr(content, "text"):
                        content_parts.append(content.text)
                    elif hasattr(content, "data"):
                        content_parts.append(str(content.data))

                return "\n".join(content_parts) if content_parts else "Tool call completed"
            else:
                return "Tool call completed"

        except Exception as e:
            logger.error(f"Failed to call tool {tool_name}: {e}")
            raise

    async def read_resource(self, uri: str) -> str:
        if not self.initialized:
            await self.initialize()

        resource = self.resources.get(uri)
        if not resource:
            raise ValueError(f"Resource '{uri}' not found")

        try:
            server_name = None
            for name, session in self.sessions.items():
                resources_response = await session.list_resources()
                if any(str(r.uri) == uri for r in resources_response.resources):
                    server_name = name
                    break

            if not server_name:
                raise ValueError(f"Resource '{uri}' not found on any server")

            session = self.sessions[server_name]

            uri_url = AnyUrl(uri)
            result = await session.read_resource(uri_url)

            if hasattr(result, "contents") and result.contents:
                content_parts = []
                for content in result.contents:
                    if hasattr(content, "text"):
                        content_parts.append(content.text)
                    elif hasattr(content, "data"):
                        content_parts.append(str(content.data))

                return "\n".join(content_parts) if content_parts else ""
            else:
                return ""

        except Exception as e:
            logger.error(f"Failed to read resource {uri}: {e}")
            raise

    async def get_prompt(self, prompt_name: str, arguments: dict[str, Any] | None = None) -> dict[str, Any]:
        if not self.initialized:
            await self.initialize()

        prompt = self.prompts.get(prompt_name)
        if not prompt:
            raise ValueError(f"Prompt '{prompt_name}' not found")

        try:
            server_name = None
            for name, session in self.sessions.items():
                prompts_response = await session.list_prompts()
                if any(p.name == prompt_name for p in prompts_response.prompts):
                    server_name = name
                    break

            if not server_name:
                raise ValueError(f"Prompt '{prompt_name}' not found on any server")

            session = self.sessions[server_name]

            result = await session.get_prompt(prompt_name, arguments or {})

            if hasattr(result, "messages"):
                messages = []
                for msg in result.messages:
                    content_text = ""
                    if msg.content:
                        if isinstance(msg.content, list | tuple) and len(msg.content) > 0:
                            first_content = msg.content[0]
                            if hasattr(first_content, "text"):
                                content_text = first_content.text
                        elif hasattr(msg.content, "text"):
                            content_text = msg.content.text
                    messages.append({"role": msg.role, "content": content_text})
                return {"messages": messages}
            else:
                return {}

        except Exception as e:
            logger.error(f"Failed to get prompt {prompt_name}: {e}")
            raise

    def get_servers_info(self) -> list[Any]:
        return list(self.tools.values())

    def get_resources(self) -> list[Any]:
        return list(self.resources.values())

    def get_prompts(self) -> list[Any]:
        return list(self.prompts.values())

    async def get_tools(self, tool_names: list[str] | None = None) -> list[Any]:
        if not self.initialized:
            await self.initialize()
            self._auto_initialized = True

        if tool_names:
            tool_names_lower = {name.lower() for name in tool_names}
            mcp_tools = []
            for tool_name, tool in self.tools.items():
                if tool_name.lower() in tool_names_lower:
                    mcp_tools.append(tool)
            return mcp_tools

        return list(self.tools.values())

    def to_openai_format(self) -> list[dict[str, Any]]:
        if not self.initialized:
            return []

        openai_tools = []
        for tool in self.tools.values():
            openai_tool = {
                "type": "function",
                "function": {
                    "name": tool.name,
                    "description": tool.description or "",
                    "parameters": tool.input_schema or {},
                },
            }
            openai_tools.append(openai_tool)

        return openai_tools

    def format_tools_for_prompt(self) -> str:
        if not self.initialized:
            return "MCP tools will be available after initialization"

        if not self.tools:
            return "No MCP tools available"

        tools_info = []
        for tool in self.tools.values():
            tool_info = f"- {tool.name}: {tool.description or 'No description'}"
            if hasattr(tool, "input_schema") and tool.input_schema and "properties" in tool.input_schema:
                params = ", ".join(tool.input_schema["properties"].keys())
                tool_info += f" (parameters: {params})"
            tools_info.append(tool_info)

        return "\n".join(tools_info)

    async def disconnect(self):
        try:
            for server_name, session in list(self.sessions.items()):
                try:
                    if hasattr(session, "close"):
                        await session.close()
                except Exception as e:
                    logger.debug(f"Error closing session {server_name}: {e}")

            self.sessions.clear()

            try:
                await self.exit_stack.aclose()
            except Exception as e:
                if any(
                    keyword in str(e).lower()
                    for keyword in [
                        "cancel scope",
                        "taskgroup",
                        "generator didn't stop",
                        "cancellederror",
                        "generatorexit",
                        "baseexceptiongroup",
                        "unhandled errors in a taskgroup",
                    ]
                ):
                    logger.debug(f"MCP client exit stack cleanup error (suppressed): {e}")
                else:
                    logger.error(f"Error closing exit stack: {e}")

        except Exception as e:
            logger.error(f"Error during disconnect: {e}")
        finally:
            self.initialized = False

    async def __aenter__(self):
        if not self.initialized:
            await self.initialize()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.disconnect()



from dataclasses import dataclass, field
from os import getenv
from typing import Any

from hypertic.embedders.base import BaseEmbedder
from hypertic.utils.log import get_logger

logger = get_logger(__name__)

try:
    from cohere import AsyncClientV2 as CohereAsyncClient, ClientV2 as CohereClient
except ImportError as err:
    raise ImportError("`cohere` not installed. Install with: pip install cohere") from err


@dataclass
class CohereEmbedder(BaseEmbedder):
    api_key: str | None = None
    model: str = "embed-english-v3.0"
    input_type: str = "search_document"
    truncate: str | None = None

    client: Any | None = field(default=None, init=False)
    async_client: Any | None = field(default=None, init=False)

    def __post_init__(self):
        self.api_key = self.api_key or getenv("COHERE_API_KEY")

    async def initialize(self) -> bool:
        try:
            _ = self._get_async_client()
            return True
        except Exception as e:
            logger.error(f"Error initializing Cohere client: {e}", exc_info=True)
            return False

    def _get_client(self):
        if self.client is None:
            if self.api_key is not None:
                self.client = CohereClient(api_key=self.api_key)
            else:
                self.client = CohereClient()
        return self.client

    def _get_async_client(self):
        if self.async_client is None:
            if self.api_key is not None:
                self.async_client = CohereAsyncClient(api_key=self.api_key)
            else:
                self.async_client = CohereAsyncClient()
        return self.async_client

    def _get_request_params(self, input_data):
        params = {
            "texts": input_data if isinstance(input_data, list) else [input_data],
            "model": self.model,
            "input_type": self.input_type,
        }

        if self.truncate:
            params["truncate"] = self.truncate

        return params

    async def embed(self, text: str) -> list[float]:
        try:
            client = self._get_async_client()
            response = await client.embed(**self._get_request_params(text))
            if response.embeddings and response.embeddings.float_:
                embedding = response.embeddings.float_[0]
                return list(embedding) if embedding else []
            return []
        except Exception as e:
            logger.error(f"Error creating embedding: {e}", exc_info=True)
            return []

    async def embed_batch(self, texts: list[str]) -> list[list[float]]:
        try:
            client = self._get_async_client()
            response = await client.embed(**self._get_request_params(texts))
            if response.embeddings and response.embeddings.float_:
                embeddings = response.embeddings.float_
                return [list(emb) for emb in embeddings] if embeddings else []
            return []
        except Exception as e:
            logger.error(f"Error creating batch embeddings: {e}", exc_info=True)
            return []

    def initialize_sync(self) -> bool:
        try:
            _ = self._get_client()
            return True
        except Exception as e:
            logger.error(f"Error initializing Cohere client: {e}", exc_info=True)
            return False

    def embed_sync(self, text: str) -> list[float]:
        try:
            client = self._get_client()
            response = client.embed(**self._get_request_params(text))
            if response.embeddings and response.embeddings.float_:
                embedding = response.embeddings.float_[0]
                return list(embedding) if embedding else []
            return []
        except Exception as e:
            logger.error(f"Error creating embedding: {e}", exc_info=True)
            return []

    def embed_batch_sync(self, texts: list[str]) -> list[list[float]]:
        try:
            client = self._get_client()
            response = client.embed(**self._get_request_params(texts))
            if response.embeddings and response.embeddings.float_:
                embeddings = response.embeddings.float_
                return [list(emb) for emb in embeddings] if embeddings else []
            return []
        except Exception as e:
            logger.error(f"Error creating batch embeddings: {e}", exc_info=True)
            return []



import pytest

from hypertic.utils.files import File, FileProcessor, FileType


@pytest.mark.unit
class TestFileProcessor:
    def test_file_processor_process_message_no_files(self):
        """Test processing message without files."""
        message = {"role": "user", "content": "Hello"}
        result = FileProcessor.process_message(message, "OpenAI")
        assert result == message

    def test_file_processor_process_message_with_files(self):
        """Test processing message with files."""
        message = {
            "role": "user",
            "content": "Hello",
            "files": ["path/to/image.jpg"],
        }
        assert "files" in message

    def test_file_processor_process_files(self, temp_dir):
        """Test processing files from directory."""
        import os

        test_file = os.path.join(temp_dir, "test.txt")
        with open(test_file, "w") as f:
            f.write("test content")
        files = FileProcessor.process_files([test_file])
        assert len(files) == 1
        assert files[0].filepath == test_file

    @pytest.mark.parametrize(
        "provider",
        ["OpenAI", "Anthropic", "Google", "Groq"],
    )
    def test_file_processor_process_message_providers(self, provider):
        """Test processing message with different providers."""
        message = {"role": "user", "content": "Hello"}
        result = FileProcessor.process_message(message, provider)
        assert result == message


@pytest.mark.unit
class TestFile:
    def test_file_creation(self):
        """Test file creation with auto-detection."""
        file_obj = File(filepath="path/to/file.jpg")
        assert file_obj.filepath == "path/to/file.jpg"

    def test_file_with_type(self):
        """Test file creation with explicit type."""
        file_obj = File(filepath="path/to/file.jpg", file_type=FileType.IMAGE)
        assert file_obj.file_type == FileType.IMAGE


@pytest.mark.unit
class TestFileType:
    @pytest.mark.parametrize(
        "file_type",
        [FileType.IMAGE, FileType.AUDIO, FileType.DOCUMENT, FileType.VIDEO, FileType.FILE],
    )
    def test_file_type_values(self, file_type):
        """Test that all file types are valid enum members."""
        assert file_type in FileType



import json
from os import getenv
from typing import Literal, Optional

from pydantic import model_validator

from hypertic.tools.base import BaseToolkit, tool

try:
    from openai import OpenAI
except ImportError as err:
    raise ImportError("openai library not installed. Install with: pip install openai") from err


class DalleTools(BaseToolkit):
    api_key: Optional[str] = None
    model: str = "dall-e-3"
    size: Optional[Literal["256x256", "512x512", "1024x1024", "1792x1024", "1024x1792"]] = "1024x1024"
    quality: Literal["standard", "hd"] = "standard"
    style: Literal["vivid", "natural"] = "vivid"
    n: int = 1

    @model_validator(mode="after")
    def _check_api_key(self):
        if not self.api_key:
            self.api_key = getenv("OPENAI_API_KEY")
        return self

    @tool
    def generate(self, query: str) -> str:
        """Generate an image from a text query using DALL-E.

        Args:
            query: The text description of the image to generate.

        Returns:
            JSON string containing the generated image URL(s) and metadata.
        """
        if not self.api_key:
            raise ValueError("API key is required. Set it when initializing DalleTools or via OPENAI_API_KEY environment variable.")

        if self.api_key is not None:
            client = OpenAI(api_key=self.api_key)
        else:
            client = OpenAI()

        try:
            n_images = 1 if self.model == "dall-e-3" else min(self.n, 10)

            if self.model == "dall-e-3":
                response = client.images.generate(
                    model=self.model,
                    prompt=query,
                    size=self.size,
                    quality=self.quality,
                    style=self.style,
                    n=1,
                )
            else:
                response = client.images.generate(
                    model=self.model,
                    prompt=query,
                    size=self.size,
                    n=n_images,
                )

            images = []
            if response.data:
                for image in response.data:
                    images.append(
                        {
                            "url": image.url,
                            "revised_prompt": getattr(image, "revised_prompt", None),
                        }
                    )

            result = {
                "model": self.model,
                "query": query,
                "images": images,
                "count": len(images),
            }

            return json.dumps(result, indent=2)
        except Exception as e:
            raise ValueError(f"Failed to generate image: {e!s}") from e



from importlib import import_module
from typing import TYPE_CHECKING

from hypertic.tools.base import BaseToolkit, tool

if TYPE_CHECKING:
    from hypertic.tools.dalle.dalle import DalleTools
    from hypertic.tools.duckduckgo.duckduckgo import DuckDuckGoTools
    from hypertic.tools.filesystem.filesystem import FileSystemTools
    from hypertic.tools.mcp.client import ExecutableTool, MCPServers

_module_lookup = {
    "DalleTools": "hypertic.tools.dalle.dalle",
    "DuckDuckGoTools": "hypertic.tools.duckduckgo.duckduckgo",
    "FileSystemTools": "hypertic.tools.filesystem.filesystem",
    "MCPServers": "hypertic.tools.mcp.client",
    "ExecutableTool": "hypertic.tools.mcp.client",
}


def __getattr__(name: str):
    if name in _module_lookup:
        module_path = _module_lookup[name]
        module = import_module(module_path)
        return getattr(module, name)
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


__all__ = list(_module_lookup.keys()) + ["BaseToolkit", "tool"]



import pytest

from hypertic.vectordb.chunking.base import Chunk
from hypertic.vectordb.chunking.strategies import DocumentChunker


class TestChunkingIntegration:
    @pytest.fixture
    def chunker(self):
        return DocumentChunker(chunk_size=100, chunk_overlap=20)

    def test_chunking_small_text(self, chunker):
        text = "This is a short text that should fit in one chunk."
        chunks = chunker.chunk_sync(text)

        assert len(chunks) >= 1
        assert all(isinstance(chunk, Chunk) for chunk in chunks)

    def test_chunking_large_text(self, chunker):
        text = "This is a test. " * 100  # Create long text
        chunks = chunker.chunk_sync(text)

        assert len(chunks) > 1
        assert all(isinstance(chunk, Chunk) for chunk in chunks)

    def test_chunking_with_overlap(self, chunker):
        text = "This is a test. " * 50
        chunks = chunker.chunk_sync(text)

        if len(chunks) > 1:
            assert chunks[0].overlap_with_next is True
            assert chunks[-1].overlap_with_previous is True

    def test_chunking_preserves_content(self, chunker):
        text = "This is a test document with multiple sentences. " * 10
        chunks = chunker.chunk_sync(text)

        reconstructed = " ".join(chunk.content for chunk in chunks)

        assert len(reconstructed) >= len(text) or True

    @pytest.mark.asyncio
    async def test_chunking_async(self, chunker):
        text = "This is a test. " * 50
        chunks = await chunker.chunk(text)

        assert len(chunks) > 0
        assert all(isinstance(chunk, Chunk) for chunk in chunks)

    def test_chunking_different_sizes(self):
        chunker_small = DocumentChunker(chunk_size=50, chunk_overlap=10)
        chunker_large = DocumentChunker(chunk_size=200, chunk_overlap=20)

        text = "This is a test. " * 50

        chunks_small = chunker_small.chunk_sync(text)
        chunks_large = chunker_large.chunk_sync(text)

        assert len(chunks_small) >= len(chunks_large)

    def test_chunking_metadata(self, chunker):
        text = "This is a test document."
        chunks = chunker.chunk_sync(text)

        assert all(hasattr(chunk, "metadata") for chunk in chunks)

    def test_chunking_chunk_type(self, chunker):
        text = "This is a test document."
        chunks = chunker.chunk_sync(text)

        assert all(chunk.chunk_type == "document" for chunk in chunks)


class TestChunkingWithVectorDB:
    @pytest.fixture
    def chunker(self):
        return DocumentChunker(chunk_size=100, chunk_overlap=20)

    def test_chunking_for_vectordb(self, chunker, mock_embedder):
        text = "This is a test document. " * 20
        chunks = chunker.chunk_sync(text)

        assert len(chunks) > 0
        assert all(hasattr(chunk, "content") for chunk in chunks)

    def test_chunking_with_embedding(self, chunker, mock_embedder):
        text = "This is a test document."
        chunks = chunker.chunk_sync(text)

        assert len(chunks) > 0
        assert all(isinstance(chunk.content, str) for chunk in chunks)



from dataclasses import dataclass, field
from os import getenv
from typing import Any

from hypertic.embedders.base import BaseEmbedder
from hypertic.utils.log import get_logger

logger = get_logger(__name__)

try:
    from google.genai import Client as GeminiClient
except ImportError as err:
    raise ImportError("`google-genai` not installed. Install with: pip install google-genai") from err


@dataclass
class GoogleEmbedder(BaseEmbedder):
    api_key: str | None = None
    model: str = "text-embedding-004"
    task_type: str | None = None
    title: str | None = None

    client: Any | None = field(default=None, init=False)
    aio_client: Any | None = field(default=None, init=False)

    def __post_init__(self):
        self.api_key = self.api_key or getenv("GEMINI_API_KEY")

    async def initialize(self) -> bool:
        try:
            _ = self._get_aio_client()
            return True
        except Exception as e:
            logger.error(f"Error initializing Google client: {e}", exc_info=True)
            return False

    def _get_client(self):
        if self.client is None:
            if self.api_key is not None:
                self.client = GeminiClient(api_key=self.api_key)
            else:
                self.client = GeminiClient()
        return self.client

    def _get_aio_client(self):
        if self.aio_client is None:
            if self.api_key is not None:
                client = GeminiClient(api_key=self.api_key)
            else:
                client = GeminiClient()
            self.aio_client = client.aio
        return self.aio_client

    def _get_request_config(self):
        if self.task_type or self.title:
            from google.genai.types import EmbedContentConfig

            config = EmbedContentConfig()
            if self.task_type:
                config.task_type = self.task_type
            if self.title:
                config.title = self.title
            return config
        return None

    async def embed(self, text: str) -> list[float]:
        try:
            aio_client = self._get_aio_client()
            config = self._get_request_config()
            response = await aio_client.models.embed_content(model=self.model, contents=text, config=config)
            if response.embeddings and len(response.embeddings) > 0:
                return response.embeddings[0].values or []
            return []
        except Exception as e:
            logger.error(f"Error creating embedding: {e}", exc_info=True)
            return []

    async def embed_batch(self, texts: list[str]) -> list[list[float]]:
        try:
            aio_client = self._get_aio_client()
            config = self._get_request_config()
            response = await aio_client.models.embed_content(model=self.model, contents=texts, config=config)
            if response.embeddings:
                return [emb.values or [] for emb in response.embeddings]
            return []
        except Exception as e:
            logger.error(f"Error creating batch embeddings: {e}", exc_info=True)
            return []

    def initialize_sync(self) -> bool:
        try:
            _ = self._get_client()
            return True
        except Exception as e:
            logger.error(f"Error initializing Google client: {e}", exc_info=True)
            return False

    def embed_sync(self, text: str) -> list[float]:
        try:
            client = self._get_client()
            config = self._get_request_config()
            response = client.models.embed_content(model=self.model, contents=text, config=config)
            if response.embeddings and len(response.embeddings) > 0:
                return response.embeddings[0].values or []
            return []
        except Exception as e:
            logger.error(f"Error creating embedding: {e}", exc_info=True)
            return []

    def embed_batch_sync(self, texts: list[str]) -> list[list[float]]:
        try:
            client = self._get_client()
            config = self._get_request_config()
            response = client.models.embed_content(model=self.model, contents=texts, config=config)
            if response.embeddings:
                return [emb.values or [] for emb in response.embeddings]
            return []
        except Exception as e:
            logger.error(f"Error creating batch embeddings: {e}", exc_info=True)
            return []



import pytest

from hypertic.utils.files import File, FileType


@pytest.mark.unit
class TestFileType:
    @pytest.mark.parametrize(
        "file_type",
        [FileType.IMAGE, FileType.AUDIO, FileType.DOCUMENT, FileType.VIDEO, FileType.FILE],
    )
    def test_file_type_enum_members(self, file_type):
        """Test that all file types are valid enum members."""
        assert file_type in FileType

    @pytest.mark.parametrize(
        "file_type,expected_value",
        [
            (FileType.IMAGE, "image"),
            (FileType.AUDIO, "audio"),
            (FileType.DOCUMENT, "document"),
            (FileType.VIDEO, "video"),
            (FileType.FILE, "file"),
        ],
    )
    def test_file_type_values(self, file_type, expected_value):
        """Test that file types have correct string values."""
        assert file_type.value == expected_value


@pytest.mark.unit
class TestFile:
    def test_file_creation(self):
        """Test file creation with auto-detection."""
        file_obj = File(filepath="path/to/file.jpg")
        assert file_obj.filepath == "path/to/file.jpg"
        assert file_obj.file_type == FileType.IMAGE

    def test_file_with_type(self):
        """Test file creation with explicit type."""
        file_obj = File(filepath="path/to/file.jpg", file_type=FileType.IMAGE)
        assert file_obj.filepath == "path/to/file.jpg"
        assert file_obj.file_type == FileType.IMAGE

    @pytest.mark.parametrize(
        "filepath,file_type",
        [
            ("image.png", FileType.IMAGE),
            ("audio.mp3", FileType.AUDIO),
            ("doc.pdf", FileType.DOCUMENT),
            ("video.mp4", FileType.VIDEO),
        ],
    )
    def test_file_with_explicit_type(self, filepath, file_type):
        """Test file creation with explicit type override."""
        file_obj = File(filepath=filepath, file_type=file_type)
        assert file_obj.file_type == file_type
